import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random

## TODO THIS BECOME VIZ TRACKER!! MODEL-VIEW-CONTROLLER Pattern!!!  THis is the MODEL, STREAMLIT IS the VIEW, STreamlit-UI+Methods called by that are CONTROLLER
class StreamlitVizTracker():
    """Track the status of the visualizations and models loaded in the Streamlit Web View. This is the Model part of the MVC pattern"""
    _set_block_container_style()
    loaded_word_embeding_pipes = []
    loaded_sentence_embeding_pipes = [] # todo track
    loaded_document_classifier_pipes = []
    loaded_token_pipes = []
    loaded_token_level_classifiers = []
    footer_displayed = False



    @staticmethod
    def pad_duplicate_tokens(tokens):
        """For every duplicate token in input list, ads N whitespaces for the Nth duplicate"""
        duplicates = {}

        for i,s in enumerate(tokens) :
            if s in duplicates.keys():duplicates[s].append(i)
            else :                    duplicates[s]=[i]
        for i, d in enumerate(duplicates.items()):
            for i,idx in enumerate(d[1]):tokens[idx]=d[0]+' '*i
        return tokens
    @staticmethod
    def RAW_HTML_link(text,url,CSS_class):return f"""<p class="{CSS_class}" style="padding:0px;" > <a href="{url}">{text}</a> </p>"""
    @staticmethod
    def style_link(text,url, CSS_class):
        return  f"""
<p class="{CSS_class}" style="
width: fit-content;
padding:0px;
color : #1E77B7; 
font-family: 'Roboto', sans-serif;
font-weight: bold;
font-size: 14px;
line-height: 17px;
box-sizing: content-box;
overflow: hidden;
display: block;
color: #0098da !important;
word-wrap: break-word;
" >
<a href="{url}">{text}</a> 
</p>"""
    @staticmethod
    def style_model_link(model,text,url, CSS_class):
        return  f"""
<p class="{CSS_class}" style="
width: fit-content;
padding:0px;
color : #1E77B7; 
font-family: 'Roboto', sans-serif;
font-weight: bold;
font-size: 14px;
line-height: 17px;
box-sizing: content-box;
overflow: hidden;
display: block;
color: #0098da !important;
word-wrap: break-word;
" >
<a href="{url}">{text}<div style:"color=rgb(246, 51, 102);">{model}</div></a> 
</p>"""
    @staticmethod
    def display_infos():
        FOOTER       = """<span style="font-size: 0.75em">{}</span>"""
        field_info   = """**INFO:** You can type in the model selection fields to search and filter."""
        iso_info     = """**INFO:** NLU model references have the structure: `<iso_language_code>.<model_name>.<dataset>` . [Based on the `ISO Language Codes`](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes). If no language defined, `en.` will be assumed as default ' ,"""
        ISO_FOOTER   = FOOTER.format(field_info)
        FIELD_FOOTER = FOOTER.format(iso_info)
        st.sidebar.markdown(ISO_FOOTER, unsafe_allow_html=True)
        st.sidebar.markdown(FIELD_FOOTER, unsafe_allow_html=True)
    @staticmethod
    def display_footer():
        if not StreamlitVizTracker.footer_displayed:
            StreamlitVizTracker.display_infos()
            nlu_link     = 'https://nlu.johnsnowlabs.com/'
            nlp_link     = 'http://nlp.johnsnowlabs.com/'
            jsl_link     = 'https://johnsnowlabs.com/'
            doc_link     = 'TODO'
            powerd_by    = f"""Powerd by [`NLU`]({nlu_link }) & [`Spark NLP`]({nlp_link}) from [`John Snow Labs `]({jsl_link}) Checkout [`The Docs`]({doc_link}) for more infos"""
            FOOTER       = """<span style="font-size: 0.75em">{}</span>"""
            POWER_FOOTER = FOOTER.format(powerd_by)
            st.sidebar.markdown(POWER_FOOTER, unsafe_allow_html=True)
            StreamlitVizTracker.footer_displayed = True
    @staticmethod
    def show_logo(sidebar=True):
        HTML_logo = """
    <div>
      <a href="https://www.johnsnowlabs.com/">
         <img src="https://nlp.johnsnowlabs.com/assets/images/logo.png" width="300"  height="100" >
       </a>
    </div>
        """
        if sidebar : st.sidebar.markdown(HTML_logo, unsafe_allow_html=True)
        else: st.markdown(HTML_logo, unsafe_allow_html=True)
    @staticmethod
    def display_embed_vetor_information(embed_component,embed_mat):
        name = StreamlitUtilsOS.extract_name(embed_component)
        if name =='': name = 'See modelshub for more details'
        exp = st.beta_expander("Vector information")
        exp.code({"Vector Dimension ":embed_mat.shape[1],
                  "Num Vectors":embed_mat.shape[0] + embed_mat.shape[0],
                  'Vector Name':name})
    @staticmethod
    def display_model_info(model2viz=' ',pipes=[],apply_style=True, display_component_wise_info=True,display_component_summary=True):
        """Display Links to Modelhub for every NLU Ref loaded and also every component in pipe"""
        default_modelhub_link = 'https://nlp.johnsnowlabs.com/models'
        nlu_refs = model2viz.split(' ')
        # for p in classifier_pipes + embed_pipes + token_pipes  :nlu_refs.append(p.nlu_ref)

        for p in StreamlitVizTracker.loaded_word_embeding_pipes + StreamlitVizTracker.loaded_document_classifier_pipes + StreamlitVizTracker.loaded_token_pipes  : nlu_refs.append(p.nlu_ref)
        nlu_refs = set(nlu_refs)
        st.sidebar.subheader("NLU pipeline components info")
        nlu_ref_infos = []
        nlu_ref_infos.append(StreamlitVizTracker.style_link("Search over 1000+ scalable SOTA models in John Snow Labs Modelhub", default_modelhub_link, CSS_class='nlu_model_info'))
        for nlu_ref in nlu_refs :
            model_hub_link = ModelHubUtils.get_url_by_nlu_refrence(nlu_ref)
            link_text =f'JSL Modelhub page for '# {nlu_ref}'
            if model_hub_link is None or model_hub_link == default_modelhub_link :
                continue
                # link_text =f'More infos here {nlu_ref}'
                # model_hub_link = default_modelhub_link
            if apply_style:
                nlu_ref_info = StreamlitVizTracker.style_model_link(nlu_ref, link_text, model_hub_link, CSS_class='nlu_model_info')
                nlu_ref_infos.append(nlu_ref_info)
                # st.sidebar.write(VizUtilsStreamlitOS.style_link(link_text,model_hub_link),unsafe_allow_html=True)
            else :
                nlu_ref_info = StreamlitVizTracker.RAW_HTML_link(link_text, model_hub_link, CSS_class='nlu_model_info')
                nlu_ref_infos.append(nlu_ref_info)
                # st.sidebar.write(nlu_ref_info)

        n = '\n'
        HTML_INFO = f"<p>{n.join(nlu_ref_infos)}</p>"
        st.sidebar.markdown(HTML_INFO, unsafe_allow_html=True)
        c_names = []
        if display_component_wise_info :
            for pipe in pipes :
                if pipe is None : continue
                for c in pipe.components :
                    c_name = f"`{type(c.model).__name__}`"
                    c_names.append(c_name)
            if model2viz[-1]==' ': model2viz = model2viz[:-1]
        FOOTER       = """<span style="font-size: 0.75em">{}</span>"""
        component_info = f"**Info:** You can load all models active in 1 line via `nlu.load('{' '.join(nlu_refs)}')` which provides this with a optimized CPU build and components: {', '.join(set(c_names))}"
        component_info = FOOTER.format(component_info)
        st.sidebar.markdown(component_info, unsafe_allow_html=True)
        if display_component_summary:
            parameter_infos = {}
            for p in pipes :
                if p is None : continue
                parameter_infos[p.nlu_ref]=StreamlitVizTracker.get_pipe_param_dict(p)
            exp = st.beta_expander("NLU Pipeline components and parameters information")
            exp.write(parameter_infos)
    @staticmethod
    def get_pipe_param_dict(pipe):
        # loop over ever model in pipeline stages  and then loop over the models params
        all_params = {}
        from sparknlp.base import LightPipeline
        stages = pipe.spark_transformer_pipe.pipeline_model.stages if isinstance(pipe.spark_transformer_pipe, (LightPipeline)) else pipe.spark_transformer_pipe.stages
        for stage in stages:
            all_params[str(stage)]={}
            params = stage.extractParamMap()
            for param_name, param_value in params.items():
                # print(f'model={stage} param_name={param_name}, param_value={param_value}')
                all_params[str(stage)][param_name.name]=param_value

        return all_params



    # @staticmethod
    # def viz_streamlit(
    #         pipe,
    #         # Base Params
    #         text:Union[str, List[str], pd.DataFrame, pd.Series],
    #         model_selection:List[str]=[],
    #         # NER PARAMS
    #         # default_ner_model2viz:Union[str, List[str]] = 'en.ner.onto.electra.base',
    #         # SIMILARITY PARAMS
    #         similarity_texts:Tuple[str,str]= ('I love NLU <3', 'I love Streamlit <3'),
    #         title:str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code' ,
    #         sub_title:str = 'Play with over 1000+ scalable enterprise NLP models',
    #         side_info:str = None,
    #         # UI PARAMS
    #         visualizers:List[str] = ( "dependency_tree", "ner",  "similarity", "token_features", 'classification','manifold'),
    #         show_models_info:bool = True,
    #         show_model_select:bool = True,
    #         show_viz_selection:bool = False,
    #         show_logo:bool=True,
    #         set_wide_layout_CSS:bool=True,
    #         show_code_snippets:bool=False,
    #         model_select_position:str = 'side' , # main or side
    #         display_infos:bool=True,
    #         key:str = "NLU_streamlit",
    #         display_footer :bool =  True ,
    #         num_similarity_cols:int=2,
    #
    #         # NEW PARAMS
    #         # MANIfold
    #         num_manifold_cols:int=3,
    #         manifold_algos:List[str]=('TSNE'),
    #
    #         # SIMY
    #         similarity_algos:List[str]=('COSINE'),
    # )-> None:
    #     """Visualize either individual building blocks for streamlit or a full UI to experiment and explore models with"""
    #     StreamlitVizTracker.footer_displayed = not display_footer
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title: st.title(title)
    #     if sub_title: st.subheader(sub_title)
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if side_info : st.sidebar.markdown(side_info)
    #     text    = st.text_area("Enter text you want to visualize below", text, key=key)
    #     ner_model_2_viz     = pipe.nlu_ref
    #     if show_model_select :
    #         show_code_snippets = st.sidebar.checkbox('Generate code snippets', value=show_code_snippets)
    #         if model_selection == [] : model_selection = Discoverer.get_components('ner',include_pipes=True)
    #         model_selection.sort()
    #         if model_select_position == 'side':ner_model_2_viz = st.sidebar.selectbox("Select a NER model.",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
    #         else : ner_model_2_viz = st.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
    #
    #     active_visualizers = visualizers
    #     if show_viz_selection: active_visualizers = st.sidebar.multiselect("Visualizers",options=visualizers,default=visualizers,key=key)
    #
    #     all_models = ner_model_2_viz + ' en.dep.typed '  if 'dependency_tree' in active_visualizers  else ner_model_2_viz
    #     ner_pipe, tree_pipe =  None,None
    #     if 'ner' in active_visualizers :
    #         ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #         StreamlitVizTracker.visualize_ner(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, show_model_select=False, show_text_input=True, show_logo=False, show_infos=False)
    #     if 'dependency_tree' in active_visualizers :
    #         tree_pipe = StreamlitUtilsOS.get_pipe('en.dep.typed') # if not ValidateVizPipe.viz_tree_satisfied(pipe) else pipe
    #         StreamlitVizTracker.visualize_dep_tree(tree_pipe, text, generate_code_sample=show_code_snippets, key=key, show_infos=False, show_logo=False)
    #     if 'token_features' in active_visualizers:
    #         ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #         StreamlitVizTracker.visualize_tokens_information(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, model_select_position=model_select_position, show_infos=False, show_logo=False, )
    #     if 'classification' in active_visualizers:
    #         ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #         StreamlitVizTracker.visualize_classes(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, model_select_position=model_select_position, show_infos=False, show_logo=False)
    #     if 'similarity' in active_visualizers:
    #         ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #         StreamlitVizTracker.display_word_similarity(ner_pipe, similarity_texts, generate_code_sample=show_code_snippets, model_select_position=model_select_position, show_infos=False, show_logo=False, num_cols=num_similarity_cols)
    #     if 'manifold' in active_visualizers :
    #         ner_pipe = pipe if ner_model_2_viz in pipe.nlu_ref.split(' ')  else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #         StreamlitVizTracker.display_low_dim_embed_viz_token(ner_pipe, similarity_texts, generate_code_sample=show_code_snippets, model_select_position=model_select_position, show_infos=False, show_logo=False, num_cols=num_manifold_cols)
    #
    #     StreamlitVizTracker.display_model_info(all_models, [ner_pipe, tree_pipe])
    #     if show_models_info :
    #         pass
    #     if display_infos : StreamlitVizTracker.display_footer()
    #
    #
    # @staticmethod
    # def visualize_classes(
    #         pipe, # nlu pipe
    #         text:Union[str,list,pd.DataFrame, pd.Series, List[str]]=('I love NLU and Streamlit and sunny days!', 'I hate rainy daiys','CALL NOW AND WIN 1000$M'),
    #         output_level:Optional[str]='document',
    #         title: Optional[str] = "Text Classification",
    #         sub_title: Optional[str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
    #         metadata : bool = False,
    #         positions : bool = False,
    #         set_wide_layout_CSS:bool=True,
    #         generate_code_sample:bool = False,
    #         key:str = "NLU_streamlit",
    #         show_model_selector : bool = True ,
    #         model_select_position:str = 'side' ,
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    #         )->None:
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title:st.header(title)
    #     if sub_title:st.subheader(sub_title)
    #
    #     # if generate_code_sample: st.code(get_code_for_viz('CLASSES',StreamlitUtilsOS.extract_name(pipe),text))
    #     if not isinstance(text, (pd.DataFrame, pd.Series)):
    #         text = st.text_area('Enter N texts, seperated by new lines to view classification results for','\n'.join(text) if isinstance(text,list) else text, key=key)
    #         text = text.split("\n")
    #         while '' in text : text.remove('')
    #     classifier_pipes = [pipe]
    #     classifier_components_usable = [e for e in Discoverer.get_components('classify',True, include_aliases=True)]
    #     classifier_components = StreamlitUtilsOS.find_all_classifier_components(pipe)
    #     loaded_classifier_nlu_refs = [c.info.nlu_ref for c in classifier_components]
    #
    #     for l in loaded_classifier_nlu_refs:
    #         if 'converter' in l :
    #             loaded_classifier_nlu_refs.remove(l)
    #             continue
    #         if l not in classifier_components_usable : classifier_components_usable.append(l)
    #     classifier_components_usable.sort()
    #     loaded_classifier_nlu_refs.sort()
    #     if show_model_selector :
    #         if model_select_position =='side':classifier_components_selection   = st.sidebar.multiselect("Pick additional Classifiers",options=classifier_components_usable,default=loaded_classifier_nlu_refs,key = key)
    #         else:classifier_components_selection   = st.multiselect("Pick additional Classifiers",options=classifier_components_usable,default=loaded_classifier_nlu_refs,key = key)
    #     # else : ValueError("Please define model_select_position as main or side")
    #     classifier_algos_to_load = list(set(classifier_components_selection) - set(loaded_classifier_nlu_refs))
    #     for classifier in classifier_algos_to_load:classifier_pipes.append(nlu.load(classifier))
    #     StreamlitVizTracker.loaded_document_classifier_pipes+= classifier_pipes
    #     if generate_code_sample:st.code(get_code_for_viz('CLASSES',[StreamlitUtilsOS.extract_name(p) for p  in classifier_pipes],text))
    #
    #     dfs = []
    #     all_classifier_cols=[]
    #     for p in classifier_pipes :
    #         df = p.predict(text, output_level=output_level, metadata=metadata, positions=positions)
    #         classifier_cols = StreamlitUtilsOS.get_classifier_cols(p)
    #         for c in classifier_cols :
    #             if c not in df.columns : classifier_cols.remove(c)
    #
    #         if 'text' in df.columns: classifier_cols += ['text']
    #         elif 'document' in df.columns: classifier_cols += ['document']
    #         all_classifier_cols+= classifier_cols
    #         dfs.append(df)
    #     df = pd.concat(dfs, axis=1)
    #     df = df.loc[:,~df.columns.duplicated()]
    #     for c in all_classifier_cols :
    #         if c not in df.columns : all_classifier_cols.remove(c)
    #     all_classifier_cols = list(set(all_classifier_cols))
    #
    #     if len(all_classifier_cols) == 0: st.error('No classes detected')
    #     else :st.write(df[all_classifier_cols],key=key)
    #     if show_infos :
    #         # VizUtilsStreamlitOS.display_infos()
    #         StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
    #         StreamlitVizTracker.display_footer()
    #
    #
    #
    # @staticmethod
    # def visualize_tokens_information(
    #         pipe, # nlu pipe
    #         text:str,
    #         title: Optional[str] = "Token Features",
    #         sub_title: Optional[str] ='Pick from `over 1000+ models` on the left and `view the generated features`',
    #         show_feature_select:bool =True,
    #         features:Optional[List[str]] = None,
    #         full_metadata: bool = True,
    #         output_level:str = 'token',
    #         positions:bool = False,
    #         set_wide_layout_CSS:bool=True,
    #         generate_code_sample:bool = False,
    #         key = "NLU_streamlit",
    #         show_model_select = True,
    #         model_select_position:str = 'side' , # main or side
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    #         show_text_input:bool = True,
    # ) -> None:
    #     """Visualizer for token features."""
    #     StreamlitVizTracker.footer_displayed=False
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title:st.header(title)
    #     # if generate_code_sample: st.code(get_code_for_viz('TOKEN',StreamlitUtilsOS.extract_name(pipe),text))
    #     if sub_title:st.subheader(sub_title)
    #     token_pipes = [pipe]
    #     if show_text_input : text = st.text_area("Enter text you want to view token features for", text, key=key)
    #     if show_model_select :
    #         token_pipes_components_usable = [e for e in Discoverer.get_components(get_all=True)]
    #         loaded_nlu_refs = [c.info.nlu_ref for c in pipe.components]
    #
    #         for l in loaded_nlu_refs:
    #             if 'converter' in l :
    #                 loaded_nlu_refs.remove(l)
    #                 continue
    #             if l not in token_pipes_components_usable : token_pipes_components_usable.append(l)
    #         token_pipes_components_usable = list(set(token_pipes_components_usable))
    #         loaded_nlu_refs = list(set(loaded_nlu_refs))
    #         if '' in loaded_nlu_refs : loaded_nlu_refs.remove('')
    #         if ' ' in loaded_nlu_refs : loaded_nlu_refs.remove(' ')
    #         token_pipes_components_usable.sort()
    #         loaded_nlu_refs.sort()
    #         if model_select_position =='side':model_selection   = st.sidebar.multiselect("Pick any additional models for token features",options=token_pipes_components_usable,default=loaded_nlu_refs,key = key)
    #         else:model_selection   = st.multiselect("Pick any additional models for token features",options=token_pipes_components_usable,default=loaded_nlu_refs,key = key)
    #         # else : ValueError("Please define model_select_position as main or side")
    #         models_to_load = list(set(model_selection) - set(loaded_nlu_refs))
    #         for model in models_to_load:token_pipes.append(nlu.load(model))
    #         StreamlitVizTracker.loaded_token_pipes+= token_pipes
    #     if generate_code_sample:st.code(get_code_for_viz('TOKEN',[StreamlitUtilsOS.extract_name(p) for p  in token_pipes],text))
    #     dfs = []
    #     for p in token_pipes:
    #         df = p.predict(text, output_level=output_level, metadata=full_metadata,positions=positions)
    #         dfs.append(df)
    #
    #
    #     df = pd.concat(dfs,axis=1)
    #     df = df.loc[:,~df.columns.duplicated()]
    #     if show_feature_select :
    #         exp = st.beta_expander("Select token features to display")
    #         features = exp.multiselect(
    #             "Token features",
    #             options=list(df.columns),
    #             default=list(df.columns)
    #         )
    #     st.dataframe(df[features])
    #     if show_infos :
    #         # VizUtilsStreamlitOS.display_infos()
    #         StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
    #         StreamlitVizTracker.display_footer()
    #
    #
    # @staticmethod
    # def visualize_dep_tree(
    #         pipe, #nlu pipe
    #         text:str = 'Billy likes to swim',
    #         title: Optional[str] = "Dependency Parse & Part-of-speech tags",
    #         sub_title: Optional[str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
    #         set_wide_layout_CSS:bool=True,
    #         generate_code_sample:bool = False,
    #         key = "NLU_streamlit",
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    #         show_text_input:bool = True,
    # ):
    #     StreamlitVizTracker.footer_displayed=False
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title:st.header(title)
    #     if show_text_input : text = st.text_area("Enter text you want to visualize dependency tree for ", text, key=key)
    #     if sub_title:st.subheader(sub_title)
    #     if generate_code_sample: st.code(get_code_for_viz('TREE',StreamlitUtilsOS.extract_name(pipe),text))
    #     pipe.viz(text,write_to_streamlit=True,viz_type='dep', streamlit_key=key)
    #     if show_infos :
    #         # VizUtilsStreamlitOS.display_infos()
    #         StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
    #         StreamlitVizTracker.display_footer()
    #
    #
    #
    # @staticmethod
    # def visualize_ner(
    #         pipe, # Nlu pipe
    #         text:str,
    #         ner_tags: Optional[List[str]] = None,
    #         show_label_select: bool = True,
    #         show_table: bool = False,
    #         title: Optional[str] = "Named Entities",
    #         sub_title: Optional[str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
    #         colors: Dict[str, str] = {},
    #         show_color_selector: bool = False,
    #         set_wide_layout_CSS:bool=True,
    #         generate_code_sample:bool = False,
    #         key = "NLU_streamlit",
    #         model_select_position:str = 'side',
    #         show_model_select : bool = True,
    #         show_text_input:bool = True,
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    #
    # ):
    #     StreamlitVizTracker.footer_displayed=False
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if show_model_select :
    #         model_selection = Discoverer.get_components('ner',include_pipes=True)
    #         model_selection.sort()
    #         if model_select_position == 'side':ner_model_2_viz = st.sidebar.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
    #         else : ner_model_2_viz = st.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
    #         pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
    #     if title: st.header(title)
    #     if show_text_input : text = st.text_area("Enter text you want to visualize NER classes for below", text, key=key)
    #     if sub_title : st.subheader(sub_title)
    #     if generate_code_sample: st.code(get_code_for_viz('NER',StreamlitUtilsOS.extract_name(pipe),text))
    #     if ner_tags is None: ner_tags = StreamlitUtilsOS.get_NER_tags_in_pipe(pipe)
    #
    #     if not show_color_selector :
    #         if show_label_select:
    #             exp = st.beta_expander("Select entity labels to highlight")
    #             label_select = exp.multiselect(
    #                 "These labels are predicted by the NER model. Select which ones you want to display",
    #                 options=ner_tags,default=list(ner_tags))
    #         else : label_select = ner_tags
    #         pipe.viz(text,write_to_streamlit=True, viz_type='ner',labels_to_viz=label_select,viz_colors=colors, streamlit_key=key)
    #     else : # TODO WIP color select
    #         cols = st.beta_columns(3)
    #         exp = cols[0].beta_expander("Select entity labels to display")
    #         color = st.color_picker('Pick A Color', '#00f900',key = key)
    #         color = cols[2].color_picker('Pick A Color for a specific entity label', '#00f900',key = key)
    #         tag2color = cols[1].selectbox('Pick a ner tag to color', ner_tags,key = key)
    #         colors[tag2color]=color
    #     if show_table : st.write(pipe.predict(text, output_level='chunk'),key = key)
    #
    #     if show_infos :
    #         # VizUtilsStreamlitOS.display_infos()
    #         StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
    #         StreamlitVizTracker.display_footer()
    #
    # @staticmethod
    # def display_word_similarity(
    #         pipe, #nlu pipe
    #         default_texts: Tuple[str, str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
    #         threshold: float = 0.5,
    #         title: Optional[str] = "Embeddings Similarity Matrix &  Visualizations  ",
    #         sub_tile :Optional[str]="Visualize `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
    #         write_raw_pandas : bool = False,
    #         display_embed_information:bool = True,
    #         similarity_matrix = True,
    #         show_algo_select : bool = True,
    #         dist_metrics:List[str]  =('cosine'),
    #         set_wide_layout_CSS:bool=True,
    #         generate_code_sample:bool = False,
    #         key:str = "NLU_streamlit",
    #         num_cols:int=2,
    #         display_scalar_similarities : bool = False ,
    #         display_similarity_summary:bool = False,
    #         model_select_position:str = 'side' , # main or side
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    # ):
    #
    #     """We visualize the following cases :
    #     1. Simmilarity between 2 words - > sim (word_emb1, word_emb2)
    #     2. Simmilarity between 2 sentences -> let weTW stand word word_emb of token T and sentence S
    #         2.1. Raw token level with merged embeddings -> sim([we11,we21,weT1], [we12,we22,weT2])
    #         2.2  Autogenerate sentemb, basically does 2.1 in the Spark NLP backend
    #         2.3 Already using sentence_embedder model -> sim(se1,se2)
    #     3. Simmilarity between token and sentence -> sim([we11,w21,wT1], se2)
    #     4. Mirrored 3
    #      """
    #     # https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics.pairwise
    #     StreamlitVizTracker.footer_displayed=False
    #     try :
    #         import plotly.express as px
    #         from sklearn.metrics.pairwise import distance_metrics
    #     except :st.error("You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title:st.header(title)
    #     if show_logo :StreamlitVizTracker.show_logo()
    #     if sub_tile : st.subheader(sub_tile)
    #
    #     StreamlitVizTracker.loaded_word_embeding_pipes = []
    #     dist_metric_algos =distance_metrics()
    #     dist_algos = list(dist_metric_algos.keys())
    #     # TODO NORMALIZE DISTANCES TO [0,1] for non cosine
    #     if 'haversine'   in dist_algos    : dist_algos.remove('haversine') # not applicable in >2D
    #     if 'precomputed' in dist_algos  : dist_algos.remove('precomputed') # Not a dist
    #     cols = st.beta_columns(2)
    #     text1 = cols[0].text_input("Text or word1",default_texts[0],key = key)
    #     text2 = cols[1].text_input("Text or word2",default_texts[1], key=key) if len(default_texts) >1  else cols[1].text_input("Text or word2",'Please enter second string',key = key)
    #     # exp = st.sidebar.beta_expander("Select additional Embedding Models and distance metric to compare ")
    #     e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)
    #     embed_algos_to_load = []
    #     embed_pipes = [pipe]
    #     dist_algo_selection = dist_metrics
    #     if show_algo_select :
    #         # emb_components_usable = Discoverer.get_components('embed')
    #         emb_components_usable = [e for e in Discoverer.get_components('embed',True, include_aliases=True) if 'chunk' not in e and 'sentence' not in e]
    #         loaded_embed_nlu_refs = []
    #         loaded_storage_refs = []
    #         for c in e_coms :
    #             if not  hasattr(c.info,'nlu_ref'): continue
    #             r = c.info.nlu_ref
    #             if 'en.' not in r and 'embed.' not  in r and 'ner' not in r : loaded_embed_nlu_refs.append('en.embed.' + r)
    #             elif 'en.'  in r and 'embed.' not  in r  and 'ner' not in r:
    #                 r = r.split('en.')[0]
    #                 loaded_embed_nlu_refs.append('en.embed.' + r)
    #             else :
    #                 loaded_embed_nlu_refs.append(StorageRefUtils.extract_storage_ref(c))
    #             loaded_storage_refs.append(StorageRefUtils.extract_storage_ref(c))
    #
    #         for l in loaded_embed_nlu_refs:
    #             if l not in emb_components_usable : emb_components_usable.append(l)
    #         # embed_algo_selection = exp.multiselect("Click to pick additional Embedding Algorithm",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
    #         # dist_algo_selection = exp.multiselect("Click to pick additional Distance Metric", options=dist_algos, default=dist_metrics, key = key)
    #         emb_components_usable.sort()
    #         loaded_embed_nlu_refs.sort()
    #         dist_algos.sort()
    #         # dist_metrics.sort()
    #         if model_select_position =='side':
    #             embed_algo_selection   = st.sidebar.multiselect("Pick additional Word Embeddings for the Similarity Matrix",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
    #             dist_algo_selection =  st.sidebar.multiselect("Pick additional Similarity Metrics ", options=dist_algos, default=dist_metrics, key = key)
    #         else :
    #             exp = st.beta_expander("Pick additional Word Embeddings and Similarity Metrics")
    #             embed_algo_selection   = exp.multiselect("Pick additional Word Embeddings for the Similarity Matrix",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
    #             dist_algo_selection =  exp.multiselect("Pick additional Similarity Metrics ", options=dist_algos, default=dist_metrics, key = key)
    #         embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))
    #
    #     for embedder in embed_algos_to_load:embed_pipes.append(nlu.load(embedder))
    #
    #     if generate_code_sample:st.code(get_code_for_viz('SIMILARITY',[StreamlitUtilsOS.extract_name(p) for p  in embed_pipes],default_texts))
    #
    #     StreamlitVizTracker.loaded_word_embeding_pipes+=embed_pipes
    #     similarity_metrics = {}
    #     embed_vector_info = {}
    #     cols_full = True
    #     col_index=0
    #     for p in embed_pipes :
    #         data1 = p.predict(text1,output_level='token').dropna()
    #         data2 = p.predict(text2,output_level='token').dropna()
    #         e_coms = StreamlitUtilsOS.find_all_embed_components(p)
    #         modelhub_links = [ModelHubUtils.get_url_by_nlu_refrence(c.info.nlu_ref) if hasattr(c.info,'nlu_ref') else ModelHubUtils.get_url_by_nlu_refrence('') for c in e_coms]
    #         e_cols = StreamlitUtilsOS.get_embed_cols(p)
    #         for num_emb,e_col in enumerate(e_cols):
    #             if col_index == num_cols-1 :cols_full=True
    #             if cols_full :
    #                 cols = st.beta_columns(num_cols)
    #                 col_index = 0
    #                 cols_full = False
    #             else:col_index+=1
    #             tok1 = data1['token']
    #             tok2 = data2['token']
    #             emb1 = data1[e_col]
    #             emb2 = data2[e_col]
    #             embed_mat1 = np.array([x for x in emb1])
    #             embed_mat2 = np.array([x for x in emb2])
    #             # e_name = e_col.split('word_embedding_')[-1]
    #             e_name = e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else e_col.split('word_embedding_')[-1] if 'en.' in e_col else e_col
    #             e_name = e_name.split('embed.')[-1] if 'en.' in e_name else e_name
    #             if 'ner' in e_name : e_name = loaded_storage_refs[num_emb]
    #
    #             embed_vector_info[e_name]= {"Vector Dimension ":embed_mat1.shape[1],
    #                                         "Num Vectors":embed_mat1.shape[0] + embed_mat1.shape[0],
    #                                         "NLU_reference":e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else ' ',
    #                                         "Spark_NLP_reference":ModelHubUtils.NLU_ref_to_NLP_ref(e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else ' '),
    #                                         "Storage Reference":loaded_storage_refs[num_emb],
    #                                         'Modelhub info': modelhub_links[num_emb]}
    #             for dist_algo in dist_algo_selection:
    #                 # scalar_similarities[e_col][dist_algo]={}
    #                 sim_score = dist_metric_algos[dist_algo](embed_mat1,embed_mat2)
    #                 sim_score = pd.DataFrame(sim_score)
    #                 sim_score.index   = tok1.values
    #                 sim_score.columns = tok2.values
    #                 sim_score.columns = StreamlitVizTracker.pad_duplicate_tokens(list(sim_score.columns))
    #                 sim_score.index   = StreamlitVizTracker.pad_duplicate_tokens(list(sim_score.index))
    #                 if write_raw_pandas :st.write(sim_score,key = key)
    #                 if sim_score.shape == (1,1) :
    #                     sim_score = sim_score.iloc[0][0]
    #                     sim_score = round(sim_score,2)
    #                     if sim_score > threshold:
    #                         st.success(sim_score)
    #                         st.success(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
    #                         st.error('No similarity matrix for only 2 tokens. Try entering at least 1 sentences in a field')
    #                     else:
    #                         st.error(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
    #                 else :
    #                     ploty_avaiable = True
    #                     # for tok emb, sum rows and norm by rows, then sum cols and norm by cols to generate a scalar from matrix
    #                     scalar_sim_score  = np.sum((np.sum(sim_score,axis=0) / sim_score.shape[0])) / sim_score.shape[1]
    #                     scalar_sim_score = round(scalar_sim_score,2)
    #
    #                     if display_scalar_similarities:
    #                         if scalar_sim_score > threshold:st.success(f'Scalar Similarity :{scalar_sim_score} for distance metric={dist_algo}')
    #                         else: st.error(f'Scalar Similarity :{scalar_sim_score} for embedder={e_col} distance metric={dist_algo}')
    #                     if similarity_matrix:
    #                         if ploty_avaiable :
    #                             fig = px.imshow(sim_score, labels=dict(color="similarity"))#, title=f'Simmilarity Matrix for embedding_model={e_name} distance metric={dist_algo}')
    #                             # st.write(fig,key =key)
    #                             similarity_metrics[f'{e_name}_{dist_algo}_similarity']={
    #                                 'scalar_similarity' : scalar_sim_score,
    #                                 'dist_metric' : dist_algo,
    #                                 'embedding_model': e_name,
    #                                 'modelhub_info' : modelhub_links[num_emb],
    #                             }
    #                             subh = f"""Embedding-Model=`{e_name}`, Similarity-Score=`{scalar_sim_score}`,  distance metric=`{dist_algo}`"""
    #                             cols[col_index].markdown(subh)
    #                             cols[col_index].write(fig, key=key)
    #                         else : pass # todo fallback plots
    #
    #     if display_similarity_summary:
    #         exp = st.beta_expander("Similarity summary")
    #         exp.write(similarity_metrics)
    #     if display_embed_information:
    #         exp = st.beta_expander("Embedding vector information")
    #         exp.write(embed_vector_info)
    #
    #     if show_infos :
    #         # VizUtilsStreamlitOS.display_infos()
    #         StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
    #         StreamlitVizTracker.display_footer()
    #
    #
    #
    #
    # @staticmethod
    # def display_low_dim_embed_viz_token(
    #         pipe, # nlu pipe
    #         default_texts: List[str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
    #         title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
    #         sub_title: Optional[str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
    #         write_raw_pandas : bool = False ,
    #         default_applicable_algos : List[str] = ('TSNE','PCA',),
    #         applicable_algos : List[str] = ("TSNE", "PCA"),#,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
    #         target_dimensions : List[int] = (1,2,3),
    #         show_algo_select : bool = True,
    #         show_embed_select : bool = True,
    #         show_color_select: bool = True,
    #         MAX_DISPLAY_NUM:int=100,
    #         display_embed_information:bool=True,
    #         set_wide_layout_CSS:bool=True,
    #         num_cols: int = 3,
    #         model_select_position:str = 'side', # side or main
    #         key:str = "NLU_streamlit",
    #         additional_classifiers_for_coloring:List[str]=['pos', 'sentiment'],
    #         extra_NLU_models_for_hueing: List[str] = ('pos','sentiment'),
    #         generate_code_sample:bool = False,
    #         show_infos:bool = True,
    #         show_logo:bool = True,
    # ):
    #     # TODO dynamic columns infer for mouse over, TOKEN LEVEL FEATURS APPLICABLE!!!!!
    #     # NIOT CRASH [1], [a b], [ab]
    #     # todo dynamic deduct Tok vs Sent vs Doc vs Chunk embeds
    #     # todo selectable color features
    #     # todo selectable mouseover features
    #     from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
    #
    #     # VizUtilsStreamlitOS.footer_displayed=False
    #     try :
    #         import plotly.express as px
    #         from sklearn.metrics.pairwise import distance_metrics
    #     except :st.error("You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")
    #     if len(default_texts) > MAX_DISPLAY_NUM : default_texts = default_texts[:MAX_DISPLAY_NUM]
    #     if set_wide_layout_CSS : _set_block_container_style()
    #     if title:st.header(title)
    #     if sub_title:st.subheader(sub_title)
    #     # if show_logo :VizUtilsStreamlitOS.show_logo()
    #
    #     # VizUtilsStreamlitOS.loaded_word_embeding_pipes = []
    #     loaded_word_embeding_pipes = []
    #
    #
    #     data = st.text_area('Enter N texts, seperated by new lines to visualize Word Embeddings for ','\n'.join(default_texts))
    #     data = data.split("\n")
    #     while '' in data : data.remove('')
    #     if len(data)<=1:
    #         st.error("Please enter more than 2 lines of text, seperated by new lines (hit <ENTER>)")
    #         return
    #     else : algos = default_applicable_algos
    #     # TODO dynamic color inference for plotting??
    #     if show_color_select: feature_to_color_by =  st.selectbox('Feature to color plots by ',['pos','sentiment',],0)
    #     text_col = 'token'
    #     embed_algos_to_load = []
    #     embed_pipes = [pipe]
    #     e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)
    #
    #     if show_algo_select :
    #         exp = st.beta_expander("Select dimension reduction technique to apply")
    #         algos = exp.multiselect(
    #             "Reduce embedding dimensionality to something visualizable",
    #             options=("TSNE", "ISOMAP",'LLE','Spectral Embedding','MDS','PCA','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),default=applicable_algos,)
    #
    #         emb_components_usable = [e for e in Discoverer.get_components('embed',True, include_aliases=True) if 'chunk' not in e and 'sentence' not in e]
    #         loaded_embed_nlu_refs = []
    #         loaded_classifier_nlu_refs = []
    #         loaded_storage_refs = []
    #         for c in e_coms :
    #             if not  hasattr(c.info,'nlu_ref'): continue
    #             r = c.info.nlu_ref
    #             if 'en.' not in r and 'embed.' not  in r and 'ner' not in r : loaded_embed_nlu_refs.append('en.embed.' + r)
    #             elif 'en.'  in r and 'embed.' not  in r  and 'ner' not in r:
    #                 r = r.split('en.')[0]
    #                 loaded_embed_nlu_refs.append('en.embed.' + r)
    #             else :
    #                 loaded_embed_nlu_refs.append(StorageRefUtils.extract_storage_ref(c))
    #             loaded_storage_refs.append(StorageRefUtils.extract_storage_ref(c))
    #
    #         for p in StreamlitVizTracker.loaded_word_embeding_pipes : loaded_embed_nlu_refs.append(p.nlu_ref)
    #         loaded_embed_nlu_refs = list(set(loaded_embed_nlu_refs))
    #         for l in loaded_embed_nlu_refs:
    #             if l not in emb_components_usable : emb_components_usable.append(l)
    #         emb_components_usable.sort()
    #         loaded_embed_nlu_refs.sort()
    #         if model_select_position =='side':
    #             embed_algo_selection   = st.sidebar.multiselect("Pick additional Word Embeddings for the Dimension Reduction",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
    #         else :
    #             exp = st.beta_expander("Pick additional Word Embeddings")
    #             embed_algo_selection   = exp.multiselect("Pick additional Word Embeddings for the Dimension Reduction",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
    #         embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))
    #
    #     for embedder in embed_algos_to_load:embed_pipes.append(nlu.load(embedder + f' {" ".join(additional_classifiers_for_coloring)}'))
    #     StreamlitVizTracker.loaded_word_embeding_pipes+=embed_pipes
    #
    #     # TODO load/update classifier pipes
    #     for nlu_ref in additional_classifiers_for_coloring :
    #         already_loaded=False
    #         if 'pos' in nlu_ref : continue
    #             # for p in  VizUtilsStreamlitOS.loaded_document_classifier_pipes:
    #             #     if p.nlu_ref == nlu_ref : already_loaded = True
    #             # if not already_loaded : VizUtilsStreamlitOS.loaded_token_level_classifiers.append(nlu.load(nlu_ref))
    #         else :
    #             for p in  StreamlitVizTracker.loaded_document_classifier_pipes:
    #                 if p.nlu_ref == nlu_ref : already_loaded = True
    #             if not already_loaded : StreamlitVizTracker.loaded_document_classifier_pipes.append(nlu.load(nlu_ref))
    #
    #     col_index = 0
    #     cols = st.beta_columns(num_cols)
    #     def are_cols_full(): return col_index == num_cols
    #     token_feature_pipe = StreamlitUtilsOS.get_pipe('en.dep.typed')
    #     ## TODO , not all pipes have sentiment/pos etc.. models for hueing loaded....
    #     ## Lets FIRST predict with the classifiers/Token level feature generators and THEN apply embed pipe??
    #     for p in StreamlitVizTracker.loaded_word_embeding_pipes :
    #         # TODO, run all classifiers pipes. FOr Sentence/Doc level stuff, we can only use Senc/Doc/Input dependent level annotators
    #         #  TODO token features TYPED DEP/ UNTYPED DEP/ POS  ---> LOAD DEP/UNTYPED DEP/ POS and then APPEN NLU_COMPONENTS!!!!! TO EXISTING PIPE
    #         classifier_cols = []
    #
    #         for class_p in StreamlitVizTracker.loaded_document_classifier_pipes:
    #             data = class_p.predict(data, output_level='document').dropna()
    #             classifier_cols.append(StreamlitUtilsOS.get_classifier_cols(class_p))
    #
    #         p = StreamlitUtilsOS.merge_token_classifiers_with_embed_pipe(p, token_feature_pipe)
    #         predictions =   p.predict(data,output_level='token').dropna()
    #         e_col = StreamlitUtilsOS.find_embed_col(predictions)
    #         e_com = StreamlitUtilsOS.find_embed_component(p)
    #         embedder_name = StreamlitUtilsOS.extract_name(e_com)
    #         emb = predictions[e_col]
    #         mat = np.array([x for x in emb])
    #         for algo in algos :
    #             if len(mat.shape)>2 : mat =mat.reshape(len(emb),mat.shape[-1])
    #
    #             # calc reduced dimensionality with every algo
    #             #todo  try/catch block for embed failures?
    #             if 1 in target_dimensions:
    #                 low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,1).fit_transform(mat)
    #                 x = low_dim_data[:,0]
    #                 y = np.zeros(low_dim_data[:,0].shape)
    #                 tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment' : predictions.sentiment})
    #                 fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=['token','text','sentiment', 'pos'])
    #                 subh = f"""Word-Embeddings =`{embedder_name}`, Manifold-Algo =`{algo}` for `D=1`"""
    #                 cols[col_index].markdown(subh)
    #                 cols[col_index].write(fig,key=key)
    #                 col_index+=1
    #                 if are_cols_full() :
    #                     cols = st.beta_columns(num_cols)
    #                     col_index = 0
    #             if 2 in target_dimensions:
    #                 low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,2).fit_transform(mat)
    #                 x = low_dim_data[:,0]
    #                 y = low_dim_data[:,1]
    #                 tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment':predictions.sentiment, })
    #                 fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=['text'])
    #                 subh = f"""Word-Embeddings =`{embedder_name}`, Manifold-Algo =`{algo}` for `D=2`"""
    #                 cols[col_index].markdown(subh)
    #                 cols[col_index].write(fig,key=key)
    #                 # st.write(fig)
    #                 col_index+=1
    #                 if are_cols_full() :
    #                     cols = st.beta_columns(num_cols)
    #                     col_index = 0
    #             if 3 in target_dimensions:
    #                 low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,3).fit_transform(mat)
    #                 x = low_dim_data[:,0]
    #                 y = low_dim_data[:,1]
    #                 z = low_dim_data[:,2]
    #                 tsne_df =  pd.DataFrame({'x':x,'y':y,'z':z, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment':predictions.sentiment, })
    #                 fig = px.scatter_3d(tsne_df, x="x", y="y", z='z',color=feature_to_color_by, hover_data=['text'])
    #                 subh = f"""Word-Embeddings =`{embedder_name}`, Manifold-Algo =`{algo}` for `D=3`"""
    #                 cols[col_index].markdown(subh)
    #                 cols[col_index].write(fig,key=key)
    #
    #                 # st.write(fig)
    #                 col_index+=1
    #                 if are_cols_full() :
    #                     cols = st.beta_columns(num_cols)
    #                     col_index = 0
    #         # Todo fancy embed infos etc
    #         # if display_embed_information: display_embed_vetor_information(e_com,mat)
    #
    #



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random

import numpy as np
import pandas as pd
import streamlit as st
import nlu
from typing import List, Sequence, Tuple, Optional, Dict, Union, Callable
import streamlit as st
import spacy
from spacy.language import Language
from spacy import displacy
from sparknlp.annotator import  *
from nlu.pipe.utils.modelhub_utils import ModelHubUtils
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils

THE_MATRIX_ARCHITECT_SCRIPT = """Neo: Who are you?
The Architect: Hello, Neo.
Neo: Who are you?
Neo: There are only two possible explanations: either no one told me, or no one knows.
The Architect: Precisely. As you are undoubtedly gathering, the anomaly's systemic, creating fluctuations in even the most simplistic equations.The Architect: Hello, Neo.
Neo: Who are you?
Neo: There are only two possible explanations: either no one told me, or no one knows.
The Architect: Precisely. As you are undoubtedly gathering, the anomaly's systemic, creating fluctuations in even the most simplistic equations.
The Architect: Hello, Neo.
Neo: Who are you?
Neo: There are only two possible explanations: either no one told me, or no one knows.
The Architect: Precisely. As you are undoubtedly gathering, the anomaly's systemic, creating fluctuations in even the most simplistic equations.
"""
THE_MATRIX_ARCHITECT_SCRIPT2= """
Neo: There are only two possible explanations: either no one told me, or no one knows.
The Architect: Precisely. As you are undoubtedly gathering, the anomaly's systemic, creating fluctuations in even the most simplistic equations.
Once again, the responses of other Neos appear on the monitors: "You can't control me! Fuck you! I'm going to kill you! You can't make me do anything!
Neo: Choice. The problem is choice.
The scene cuts to Trinity fighting an agent, and then back to the Architect's room
The Architect: The first matrix I designed was quite naturally perfect, it was a work of art, flawless, sublime. A triumph equaled only by its monumental failure. The inevitability of its doom is as apparent to me now as a consequence of the imperfection inherent in every human being, thus I redesigned it based on your history to more accurately reflect the varying grotesqueries of your nature. However, I was again frustrated by failure. I have since come to understand that the answer eluded me because it required a lesser mind, or perhaps a mind less bound by the parameters of perfection. Thus, the answer was stumbled upon by another, an intuitive program, initially created to investigate certain aspects of the human psyche. If I am the father of the matrix, she would undoubtedly be its mother.
Neo: The Oracle.
The Architect: Please. As I was saying, she stumbled upon a solution whereby nearly 99% of all test subjects accepted the program, as long as they were given a choice, even if they were only aware of the choice at a near unconscious level. While this answer functioned, it was obviously fundamentally flawed, thus creating the otherwise contradictory systemic anomaly, that if left unchecked might threaten the system itself. Ergo, those that refused the program, while a minority, if unchecked, would constitute an escalating probability of disaster.
Neo: This is about Zion.
The Architect: You are here because Zion is about to be destroyed. Its every living inhabitant terminated, its entire existence eradicated.
Neo: Bullshit.
The responses of other Neos appear on the monitors: "Bullshit!"
The Architect: Denial is the most predictable of all human responses. But, rest assured, this will be the sixth time we have destroyed it, and we have become exceedingly efficient at it.
Scene cuts to Trinity fighting an agent, and then back to the Architects room.
The Architect: The function of the One is now to return to the source, allowing a temporary dissemination of the code you carry, reinserting the prime program. After which you will be required to select from the matrix 23 individuals, 16 female, 7 male, to rebuild Zion. Failure to comply with this process will result in a cataclysmic system crash killing everyone connected to the matrix, which coupled with the extermination of Zion will ultimately result in the extinction of the entire human race.
Neo: You won't let it happen, you can't. You need human beings to survive.
The Architect: There are levels of survival we are prepared to accept. However, the relevant issue is whether or not you are ready to accept the responsibility for the death of every human being in this world.
The Architect presses a button on a pen that he is holding, and images of people from all over the matrix appear on the monitors
The Architect: It is interesting reading your reactions. Your five predecessors were by design based on a similar predication, a contingent affirmation that was meant to create a profound attachment to the rest of your species, facilitating the function of the one. While the others experienced this in a very general way, your experience is far more specific. Vis-a-vis, love.
Images of Trinity fighting the agent from Neo's dream appear on the monitors
Neo: Trinity.
The Architect: Apropos, she entered the matrix to save your life at the cost of her own.
Neo: No!
The Architect: Which brings us at last to the moment of truth, wherein the fundamental flaw is ultimately expressed, and the anomaly revealed as both beginning, and end. There are two doors. The door to your right leads to the source, and the salvation of Zion. The door to the left leads back to the matrix, to her, and to the end of your species. As you adequately put, the problem is choice. But we already know what you're going to do, don't we? Already I can see the chain reaction, the chemical precursors that signal the onset of emotion, designed specifically to overwhelm logic, and reason. An emotion that is already blinding you from the simple, and obvious truth: she is going to die, and there is nothing that you can do to stop it.
Neo walks to the door on his left
The Architect: Humph. Hope, it is the quintessential human delusion, simultaneously the source of your greatest strength, and your greatest weakness.
Neo: If I were you, I would hope that we don't meet again.
The Architect: We won't.
End Scene
"""

# nlu_streamlit.sh
# TODO NGROK INTEGRATION - 1liner for demo pyngrok
# Split releases
# ask ida for emdium posts
# NER_COLORS = {}
# TODO VIZ 4 dataset
# TODO TIME SERIES VIZ FOR DATASETS
# https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html#sklearn.decomposition.TruncatedSVD
@st.cache(allow_output_mutation=True)
def get_pipe(model='ner'): return nlu.load(model)
def visualize_side_header():
    HTML_logo = """
<div>
  <a href="https://www.johnsnowlabs.com/">
     <img src="https://nlp.johnsnowlabs.com/assets/images/logo.png" width="300"  height="100" >
   </a>
</div>
    """
    st.sidebar.markdown(HTML_logo, unsafe_allow_html=True)
def find_embed_col(df, search_multi=False):
    """Find col that contains embed"""
    if not search_multi:
        for c in df.columns:
            if 'embed'in c : return c
    else:
        e_cols =[]
        for c in df.columns:
            if 'embed'in c : e_cols.append(c)
    return  e_cols
def find_embed_component(p):
    """Find NER component in pipe"""
    for c in p.components :
        if 'embed' in c.info.outputs[0] : return c
    st.warning("No Embed model in pipe")
    return None
def display_model_info(model2viz):
    """Display Links to Modelhub for every NLU Ref loaded"""
    default_modelhub_link = 'https://modelshub.johnsnowlabs.com/'
    nlu_refs = set(model2viz.split(' '))
    for nlu_ref in nlu_refs :
        model_hub_link = ModelHubUtils.get_url_by_nlu_refrence(nlu_ref)
        if model_hub_link is not None :
            st.sidebar.write(f"[Model info for {nlu_ref}]({model_hub_link})")
        else :
            st.sidebar.write(f"[Model info for {nlu_ref}]({default_modelhub_link})")
def extract_name(component_or_pipe):
    name =''
    if hasattr(component_or_pipe.info,'nlu_ref') : name = component_or_pipe.info.nlu_ref
    elif hasattr(component_or_pipe,'storage_ref') : name = component_or_pipe.info.storage_ref
    elif hasattr(component_or_pipe,'nlp_ref') : name = component_or_pipe.info.nlp_ref
    return name
def display_embed_vetor_information(embed_component,embed_mat):
    name = extract_name(embed_component)
    if name =='': name = 'See modelshub for more details'
    exp = st.beta_expander("Vector information")
    exp.code({"Vector Dimension ":embed_mat.shape[1],
              "Num Vectors":embed_mat.shape[0] + embed_mat.shape[0],
              'Vector Name':name})
from sklearn.manifold import TSNE, Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding
from sklearn.decomposition import TruncatedSVD,DictionaryLearning, FactorAnalysis, FastICA, KernelPCA, PCA
def get_manifold_algo(algo,dim):
    # manifold
    if algo=='TSNE' : return TSNE(n_components=dim)
    if algo=='ISOMAP' : return Isomap(n_components=dim)
    if algo=='LLE' : return LocallyLinearEmbedding(n_components=dim)
    if algo=='Spectral Embedding' : return  SpectralEmbedding(n_components=dim)
    if algo=='MDS' : return MDS(n_components=dim)
    # Matrix Decomposition
    if algo== 'PCA' : return PCA(n_components=dim)
    if algo== 'SVD aka LSA' : return TruncatedSVD(n_components=dim)
    if algo =='DictionaryLearning': return DictionaryLearning(n_components=dim)
    if algo =='FactorAnalysis': return FactorAnalysis(n_components=dim)
    if algo =='FastICA': return FastICA(n_components=dim)
    if algo =='KernelPCA': return KernelPCA(n_components=dim)
    # not applicable because negative values, todo we could just take absolute values of all embeds..
    # if algo =='LatentDirichletAllocation': return LatentDirichletAllocation(n_components=dim)
    # if algo =='NMF': return NMF(n_components=dim)


import numpy as np
import plotly.express as px


def display_low_dim_embed_viz_sentence(
        pipe, # nlu pipe
        default_texts: List[str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
        title: Optional[str] = "Lower dimensional Manifold visualization for Sentence Embeddings",
        sub_title: Optional[str] = "`todo subtileme`",
        write_raw_pandas : bool = False ,
        default_applicable_algos : List[str] = ('TSNE','PCA'),
        applicable_algos : List[str] = ("TSNE", "ISOMAP",'LLE','Spectral Embedding','MDS','PCA','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
        target_dimensions : List[int] = (1,2,3),
        show_algo_select : bool = True,
        show_color_select: bool = True,
        key : Optional[str] = 'sentence_embed_manifold',
        display_embed_information:bool=True,
):
    # TODO dynamic columns infer for mouse over, sentence/doc level classifier results applicable!!!
    # todo selectable mouseover features
    # todo upload dataset

    if title:st.header(title)
    if sub_title:st.header(title)
    data = st.text_area('Enter N texts, seperated by new lines to visualize embeddings for ','\n'.join(default_texts),key=key)
    data = data.split("\n")
    while '' in data : data.remove('')
    if len(data)<=1:
        st.error("Please enter more than 2 lines of text, seperated by new lines (hit <ENTER>)")
        return

    if show_algo_select:
        exp = st.beta_expander("Select dimension reduction technique to apply")
        algos = exp.multiselect(
            "Reduce embedding dimensionality to something visualizable",
            options=applicable_algos,
            key=key,
            # default=default_applicable_algos,
        default=applicable_algos,
        )
    else : algos = default_applicable_algos
    cols = st.beta_columns(2)
    if show_color_select: feature_to_color_by = cols[0].selectbox('Feature to color plots by ',['sentiment',],0,key=key)

    detect_sentences = cols[1].checkbox("Detect sentences automatically", value=True)
    output_level = 'sentence' if detect_sentences  else 'document'
    text_col = 'sentence'
    predictions =   pipe.predict(data,output_level=output_level).dropna()
    e_col = find_embed_col(predictions)
    e_com = find_embed_component(pipe)
    embedder_name = extract_name(e_com)
    if embedder_name =='': embedder_name = 'See modelshub for more details'

    emb = predictions[e_col]
    mat = np.array([x for x in emb])
    if display_embed_information: display_embed_vetor_information(e_com,mat)
    for algo in algos :
        # calc reduced dimensionality with every algo
        if 1 in target_dimensions:
            low_dim_data = get_manifold_algo(algo,1).fit_transform(mat)
            x = low_dim_data[:,0]
            y = np.zeros(low_dim_data[:,0].shape)
            # tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col], 'sentiment':predictions.sentiment, })
            tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col],  'sentiment' : predictions.sentiment})
            fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=['text','sentiment', ],title=f'3D manifold with {algo} via {embedder_name}')
            st.write(fig,key=key)
        if 2 in target_dimensions:
            low_dim_data = get_manifold_algo(algo,2).fit_transform(mat)
            x = low_dim_data[:,0]
            y = low_dim_data[:,1]
            tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col],'sentiment':predictions.sentiment, })
            fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=['text','sentiment'],title=f'3D manifold with {algo} via {embedder_name}')
            st.write(fig,key=key)
        if 3 in target_dimensions:
            low_dim_data = get_manifold_algo(algo,3).fit_transform(mat)
            x = low_dim_data[:,0]
            y = low_dim_data[:,1]
            z = low_dim_data[:,2]
            tsne_df =  pd.DataFrame({'x':x,'y':y,'z':z, 'text':predictions[text_col],  'sentiment':predictions.sentiment, })
            fig = px.scatter_3d(tsne_df, x="x", y="y", z='z',color=feature_to_color_by, hover_data=['text','sentiment'],title=f'3D manifold with {algo} via {embedder_name}')
            st.write(fig,key=key)



def viz_streamlit(
    text:Union[str, List[str], pd.DataFrame, pd.Series], # text 2 viz
    default_ner_model2viz:Union[str, List[str]] = 'en.ner.onto.electra.base',   # models2viz
    default_sentence_embed_2viz:Union[str, List[str]] = 'en.embed_sentence.electra',
    default_word_embed_2viz:Union[str, List[str]] = 'en.embed.bert.small_L2_128 en.pos en.sentiment',
    extra_classifier_for_sentence_embedding_coloring:Union[str, List[str]] = 'en.sentiment',
    extra_classifier_for_word_embedding_coloring:Union[str, List[str]] = 'en.pos en.sentiment',
    default_manifold_text:List[str]= ('Donald Trump Likes to part', 'Angela Merkel likes to party', 'Peter HATES TO PARTTY!!!! :('),
    show_side_header:bool=True,
    show_word_embed_model_select:bool = True,
    show_sentence_embed_model_select:bool = True,
    )->None:
    """Visualize either individual building blocks for streamlit or a full UI to experiment and explore models with

    """
    st.title('NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code')

    if show_side_header :     visualize_side_header()
    if isinstance(default_ner_model2viz, List)       : default_ner_model2viz = ' '.join(default_ner_model2viz)
    if isinstance(default_sentence_embed_2viz, List)       : default_ner_model2viz = ' '.join(default_ner_model2viz)
    word_embed_2viz     = default_word_embed_2viz     +' '  +extra_classifier_for_word_embedding_coloring
    sentence_embed_2viz = default_sentence_embed_2viz +' '  +extra_classifier_for_sentence_embedding_coloring

    text    = st.text_area("Enter text you want to visualize below", text)

    #
    # if show_sentence_embed_model_select  :
    #     sentence_embed_models = nlu.get_components('embed_sentence')
    #     sentence_embed_2viz = st.sidebar.selectbox('Select a Sentence Embedding',sentence_embed_models,index=sentence_embed_models.index(default_sentence_embed_2viz.split(' ')[0]))
    # if show_word_embed_model_select  :
    #     word_embed_models = nlu.get_components('embed.')
    #     word_embed_2viz = st.sidebar.selectbox('Select a Word Embedding',word_embed_models,index=word_embed_models.index(default_word_embed_2viz.split(' ')[0]))
    # word_embed_2viz     = word_embed_2viz + ' ' + extra_classifier_for_word_embedding_coloring
    # sentence_embed_2viz = sentence_embed_2viz + ' ' + extra_classifier_for_sentence_embedding_coloring


    active_visualizers = ['embedding_sentence','embedding_word']

    if 'embedding_word' in active_visualizers :
        word_embed_pipe = get_pipe('ner sentiment pos ') # TODO PARAM pize get pipez
        display_low_dim_embed_viz_token(word_embed_pipe, default_manifold_text, applicable_algos=['TSNE','LLE'])
    # if 'embedding_sentence' in active_visualizers :
    #     sentence_embed_pipe = get_pipe(sentence_embed_2viz)
    #     display_low_dim_embed_viz_sentence(sentence_embed_pipe, default_manifold_text)
# Get a list of all NER models and define default input text
ner_models = nlu.get_components('ner',include_pipes=True)
default_text = "Sundar Pichai is the CEO of Google."


nlu.enable_streamlit_caching()
viz_streamlit(default_text,ner_models, default_manifold_text=THE_MATRIX_ARCHITECT_SCRIPT.split('\n'))



# TODO MULTI-COL-LAYOUT
# TODO Multi embed select and infer






import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class TokenFeaturesStreamlitBlock():
    @staticmethod
    def visualize_tokens_information(
            pipe, # nlu pipe
            text:str,
            title: Optional[str] = "Token Features",
            sub_title: Optional[str] ='Pick from `over 1000+ models` on the left and `view the generated features`',
            show_feature_select:bool =True,
            features:Optional[List[str]] = None,
            full_metadata: bool = True,
            output_level:str = 'token',
            positions:bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            show_model_select = True,
            model_select_position:str = 'side' , # main or side
            show_infos:bool = True,
            show_logo:bool = True,
            show_text_input:bool = True,
    ) -> None:
        """Visualizer for token features."""
        StreamlitVizTracker.footer_displayed=False
        if show_logo :StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS : _set_block_container_style()
        if title:st.header(title)
        # if generate_code_sample: st.code(get_code_for_viz('TOKEN',StreamlitUtilsOS.extract_name(pipe),text))
        if sub_title:st.subheader(sub_title)
        token_pipes = [pipe]
        if show_text_input : text = st.text_area("Enter text you want to view token features for", text, key=key)
        if show_model_select :
            token_pipes_components_usable = [e for e in Discoverer.get_components(get_all=True)]
            loaded_nlu_refs = [c.info.nlu_ref for c in pipe.components]

            for l in loaded_nlu_refs:
                if 'converter' in l :
                    loaded_nlu_refs.remove(l)
                    continue
                if l not in token_pipes_components_usable : token_pipes_components_usable.append(l)
            token_pipes_components_usable = list(set(token_pipes_components_usable))
            loaded_nlu_refs = list(set(loaded_nlu_refs))
            if '' in loaded_nlu_refs : loaded_nlu_refs.remove('')
            if ' ' in loaded_nlu_refs : loaded_nlu_refs.remove(' ')
            token_pipes_components_usable.sort()
            loaded_nlu_refs.sort()
            if model_select_position =='side':model_selection   = st.sidebar.multiselect("Pick any additional models for token features",options=token_pipes_components_usable,default=loaded_nlu_refs,key = key)
            else:model_selection   = st.multiselect("Pick any additional models for token features",options=token_pipes_components_usable,default=loaded_nlu_refs,key = key)
            # else : ValueError("Please define model_select_position as main or side")
            models_to_load = list(set(model_selection) - set(loaded_nlu_refs))
            for model in models_to_load:token_pipes.append(nlu.load(model))
            StreamlitVizTracker.loaded_token_pipes+= token_pipes
        if generate_code_sample:st.code(get_code_for_viz('TOKEN',[StreamlitUtilsOS.extract_name(p) for p  in token_pipes],text))
        dfs = []
        for p in token_pipes:
            df = p.predict(text, output_level=output_level, metadata=full_metadata,positions=positions)
            dfs.append(df)


        df = pd.concat(dfs,axis=1)
        df = df.loc[:,~df.columns.duplicated()]
        if show_feature_select :
            exp = st.beta_expander("Select token features to display")
            features = exp.multiselect(
                "Token features",
                options=list(df.columns),
                default=list(df.columns)
            )
        st.dataframe(df[features])
        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()




'''
Contains methods used to resolve a NLU reference to a NLU component.
Handler for getting default components, etcc.
'''
from sparknlp.base import Finisher,EmbeddingsFinisher
from sparknlp.annotator import SentenceEmbeddings

# <<<Name parse procedure>>>
#1.  parse NAMLE data and RETURN IT -> Detect if OC or Closed source (CS)
#2. based on wether OC or CS, use according component resolver
# 2.1 if CS_annotator, then verify licsence/ authenticate, if not do the usual (Make sure all CS imports are in seperate files)
#3. Put all components in NLU pipe and return it
#





# <<<Pythonify procedure>>>
# 1, transform DF
# 2.  Integrate outoputlevel of new annotators by getting some attriubte/str name from them.
# We cannot do isInstance() because we cannot import classes for the cmparioson
# Thus, for OUtput_level inference of singular components and the entiure pipe
# we must first check OC compoments vanilla style and if that fails we must do special infer_CS_component_level() all
# This call must infer the output level without type checks, i.e. use component infos or some string map or some trick (( !! component info!!!)

# 2. Squeeze in 9 Annotators in old extraction process, most annotators are like old ones
#
import nlu.utils.environment.authentication as auth_utils
import nlu.utils.environment.offline_load_utils as offline_utils

from pyspark.ml import PipelineModel
from sparknlp import DocumentAssembler
from sparknlp.annotator import NerConverter, MultiClassifierDLModel, PerceptronModel, ClassifierDLModel, \
    UniversalSentenceEncoder, BertEmbeddings, AlbertEmbeddings, XlnetEmbeddings, WordEmbeddingsModel, ElmoEmbeddings, \
    BertSentenceEmbeddings, TokenizerModel, SentenceDetectorDLModel, SentenceDetector, RegexMatcherModel, \
    TextMatcherModel, DateMatcher, ContextSpellCheckerModel, SymmetricDeleteModel, NorvigSweetingModel, LemmatizerModel, \
    NormalizerModel, Stemmer, NerDLModel, NerCrfModel, LanguageDetectorDL, DependencyParserModel, \
    TypedDependencyParserModel, SentimentDetectorModel, SentimentDLModel, ViveknSentimentModel, Chunker, \
    ChunkEmbeddings, StopWordsCleaner, MultiDateMatcher, T5Transformer, MarianTransformer

import nlu
#NluError, all_components_info,
from nlu import logger, Util, Embeddings, Classifier, Spellbook, ClassifierDl, \
    NLUSentenceDetector, NGram, Seq2Seq, SpellChecker, Matcher
from nlu.components import embeddings_chunker
from nlu.components.labeled_dependency_parser import LabeledDependencyParser as LabledDepParser
from nlu.components.unlabeled_dependency_parser import UnlabeledDependencyParser as UnlabledDepParser
from nlu.pipe.utils.pipe_utils import PipeUtils
from nlu.pipe.utils.component_utils import ComponentUtils

def load_offline_model(path):
    c = offline_utils.verify_and_create_model(path)
    return c
def parse_language_from_nlu_ref(nlu_ref):
    """Parse a ISO language identifier from a NLU reference which can be used to load a Spark NLP model"""
    infos = nlu_ref.split('.')
    for split in infos:
        if split in nlu.AllComponentsInfo().all_languages:
            logger.info(f'Parsed Nlu_ref={nlu_ref} as lang={split}')
            return split
    logger.info(f'Parsed Nlu_ref={nlu_ref} as lang=en')
    return 'en'




def get_default_component_of_type(missing_component_type,language='en',is_licensed=False, is_trainable_pipe = False ):
    '''
    This function returns a default component for a missing component type.
    It is used to auto complete pipelines, which are missng required components.
    These represents defaults for many applications and should be set wisely.
    :param missing_component_type: String which is either just the component type or componenttype@spark_nlp_reference which stems from a models storageref and refers to some pretrained embeddings or model
    :return: a NLU component which is a either the default if there is no '@' in the @param missing_component_type or a default component for that particualar type
    '''

    logger.info(f'Getting default for missing_component_type={ missing_component_type}')
    if not '@' in missing_component_type:
        # get default models if there is no @ in the model name included
        if missing_component_type == 'document': return Util('document_assembler', nlu_ref='document')
        if missing_component_type == 'sentence': return Util('deep_sentence_detector', nlu_ref='sentence')
        if missing_component_type == 'sentence_embeddings': return Embeddings('use',nlu_ref='embed_sentence.use')
        if 'token' in missing_component_type: return nlu.components.tokenizer.Tokenizer("default_tokenizer", nlu_ref = 'tokenize', language=language)
        if missing_component_type == 'word_embeddings': return Embeddings(annotator_class='glove', nlu_ref='embed.glove')
        if missing_component_type == 'pos':   return Classifier(nlu_ref='pos')
        if missing_component_type == 'ner':   return Classifier(nlu_ref='ner') if not is_licensed else Classifier(nlu_ref='med_ner', nlp_ref='jsl_ner_wip_clinical')
        if missing_component_type == 'ner_converter':   return Util('ner_converter', nlu_ref='entity')
        if missing_component_type == 'chunk': return nlu.chunker.Chunker(nlu_ref='chunker')
        if missing_component_type == 'ngram': return nlu.chunker.Chunker(nlu_ref='ngram')
        if missing_component_type == 'chunk_embeddings': return embeddings_chunker.EmbeddingsChunker(nlu_ref='chunk_embeddings')
        if missing_component_type == 'unlabeled_dependency': return UnlabledDepParser(nlu_ref='dep.untyped')
        if missing_component_type == 'labled_dependency': return LabledDepParser('dep',nlu_ref='dep.typed')
        if missing_component_type == 'date': return nlu.Matcher('date', nlu_ref='match.date')
        if missing_component_type == 'ner_chunk': return Util('ner_converter',nlu_ref='entitiy')
        # TODO ENTITIES REQUIREMENT: if entities is required and we are training a chunk_reoslver, we just and DOC2CHUNK!!! ??

        if missing_component_type == 'entities' and is_licensed and is_trainable_pipe: return Util('doc2chunk')
        if missing_component_type == 'entities' and is_licensed: return Util('ner_to_chunk_converter_licensed')
        if missing_component_type == 'entities': return Util('ner_converter')
        if missing_component_type == 'feature_vector': return Util('feature_assembler')

    else:
        """ These models are fetched becuase they required a storage ref, so we ad a storage ref attribute to the component info"""
        multi_lang =['ar']
        # if there is an @ in the name, we must get some specific pretrained model from the sparknlp reference that should follow after the @
        missing_component_type, storage_ref = missing_component_type.split('@')



        if 'embed' in missing_component_type:
            # TODO RESOLVE MULTI LANG EMBEDS , this should be hdled  better
            if language in multi_lang : storage_ref = resolve_multi_lang_embed(language,storage_ref)
            else :  nlu_ref,nlp_ref, is_licensed =  resolve_storage_ref(language,storage_ref,missing_component_type)

            if 'chunk_embeddings' in missing_component_type: return  set_storage_ref_and_resolution_on_component_info(embeddings_chunker.EmbeddingsChunker(nlu_ref=nlu_ref,nlp_ref=nlp_ref),storage_ref)
            else : return set_storage_ref_and_resolution_on_component_info(construct_component_from_identifier(language=language, component_type='embed', nlu_ref=nlu_ref,nlp_ref=nlp_ref, is_licensed=is_licensed), storage_ref,)

        if 'pos' in missing_component_type or 'ner' in missing_component_type:
            return construct_component_from_identifier(language=language, component_type='classifier',nlp_ref=storage_ref)
        # if 'unlabeled_dependency' in missing_component_type or 'dep.untyped' in missing_component_type:
        #     return UnlabledDepParser('dep.untyped')
        # if 'labled_dependency' in missing_component_type or 'dep.typed' in missing_component_type:
        #     return LabledDepParser('dep.typed')
        # if 'date' in missing_component_type:
        #     return None

        logger.exception("Could not resolve default component type for missing type=%s", missing_component_type)


def set_storage_ref_and_resolution_on_component_info(c,storage_ref):
    """Sets a storage ref on a components component info and returns the component """
    c.info.storage_ref = storage_ref

    return c
def resolve_storage_ref(lang, storage_ref,missing_component_type):
    """Returns a nlp_ref, nlu_ref and wether it is a licensed model or not"""
    logger.info(f"Resolving storage_ref={storage_ref} for lang={lang} and missing_component_type={missing_component_type}")
    nlu_ref,nlp_ref,is_licensed = None,None,False
    # get nlu ref


    # check if storage_ref is hardcoded
    if lang in nlu.Spellbook.licensed_storage_ref_2_nlu_ref.keys():
        if storage_ref in nlu.Spellbook.licensed_storage_ref_2_nlu_ref[lang].keys():
            nlu_ref = nlu.Spellbook.licensed_storage_ref_2_nlu_ref[lang][storage_ref]
            is_licensed = False
    elif lang in nlu.Spellbook.storage_ref_2_nlu_ref.keys():
        if storage_ref in nlu.Spellbook.storage_ref_2_nlu_ref[lang].keys():
            nlu_ref = nlu.Spellbook.storage_ref_2_nlu_ref[lang][storage_ref] # a HC model may use OS storage_ref_provider, so we dont know yet if it is licensed or not
    if nlu_ref is not None and 'xx.' in nlu_ref : lang = 'xx'

    if lang in nlu.Spellbook.pretrained_models_references.keys() and nlu_ref in nlu.Spellbook.pretrained_models_references[lang].keys():
        nlp_ref = nlu.Spellbook.pretrained_models_references[lang][nlu_ref]
    elif lang in nlu.Spellbook.pretrained_healthcare_model_references.keys() and nlu_ref in nlu.Spellbook.pretrained_healthcare_model_references[lang].keys():
        nlp_ref = nlu.Spellbook.pretrained_healthcare_model_references[lang][nlu_ref]
        is_licensed = True


    # check if storage_ref matches nlu_ref and get NLP_ref
    elif lang in nlu.Spellbook.licensed_storage_ref_2_nlu_ref.keys():
        if storage_ref in nlu.Spellbook.licensed_storage_ref_2_nlu_ref[lang].keys():
            nlu_ref = storage_ref
            nlp_ref = nlu.Spellbook.licensed_storage_ref_2_nlu_ref[lang][nlu_ref]
    elif lang in nlu.Spellbook.pretrained_models_references.keys():
        if storage_ref in nlu.Spellbook.pretrained_models_references[lang].keys():
            nlu_ref = storage_ref
            nlp_ref = nlu.Spellbook.pretrained_models_references[lang][nlu_ref]



    if nlu_ref == None and nlp_ref == None :
        logger.info("COULD NOT RESOLVE STORAGE_REF")
        if storage_ref =='' :
            if missing_component_type =='sentence_embeddings':
                logger.info("Using default storage_ref USE, assuming training mode")
                storage_ref = 'en.embed_sentence.use' # this enables default USE embeds for traianble components
                nlp_ref = 'tfhub_use'
                nlu_ref = storage_ref
            elif missing_component_type =='word_embeddings':
                logger.info("Using default storage_ref GLOVE, assuming training mode")
                storage_ref = 'en.glove' # this enables default USE embeds for traianble components
                nlp_ref = 'glove_100d'
                nlu_ref = storage_ref

        else :
            nlp_ref = storage_ref
            nlu_ref = storage_ref
        # raise  ValueError

    logger.info(f'Resolved storageref = {storage_ref} to NLU_ref = {nlu_ref} and NLP_ref = {nlp_ref}')
    return nlu_ref, nlp_ref, is_licensed



def resolve_multi_lang_embed(language,sparknlp_reference):
    """Helper Method for resolving Multi Lingual References to correct embedding"""
    if language == 'ar' and 'glove' in sparknlp_reference : return 'arabic_w2v_cc_300d'
    else : return sparknlp_reference

def  nlu_ref_to_component(nlu_reference, detect_lang=False, authenticated=False, is_recursive_call=False):
    '''
    This method implements the main namespace for all component names. It parses the input request and passes the data to a resolver method which searches the namespace for a Component for the input request
    It returns a list of NLU.component objects or just one NLU.component object alone if just one component was specified.
    It maps a correctly namespaced name to a corresponding component for pipeline
    If no lang is provided, default language eng is assumed.
    General format  <lang>.<class>.<dataset>.<embeddings>
    For embedding format : <lang>.<class>.<variant>
    This method will parse <language>.<NLU_action>
        Additional data about dataset and variant will be resolved by corrosponding action classes

    If train prefix is part of the nlu_reference ,the trainable namespace will e searched

    if 'translate_to' or 'marian' is inside of the nlu_reference, 'xx' will be prefixed to the ref and set as lang if it is not already
    Since all translate models are xx lang
    :param nlu_reference: User request (should be a NLU reference)
    :param detect_lang: Wether to automatically  detect language
    :return: Pipeline or component for the NLU reference.
    '''

    infos = nlu_reference.split('.')

    if len(infos) < 0:
        logger.exception(f"EXCEPTION: Could not create a component for nlu reference={nlu_reference}", )
        return nlu.NluError()
    language = ''
    component_type = ''
    dataset = ''
    component_embeddings = ''
    trainable = False
    resolved_component = None
    # 1. Check if either a default cmponent or one specific pretrained component or pipe  or alias of them is is requested without more sepcificatin about lang,dataset or embeding.
    # I.e. 'explain_ml' , 'explain; 'albert_xlarge_uncased' or ' tokenize'  or 'sentiment' s requested. in this case, every possible annotator must be checked.
    # format <class> or <nlu identifier>
    # type parsing via checking parts in nlu action <>


    if len(infos) == 0:
        logger.exception("Split  on query is 0.")
    # Query of format <class>, no embeds,lang or dataset specified
    elif 'train' in infos :
        trainable = True
        component_type = infos[1]
    elif 'translate_to' in nlu_reference and not 't5' in nlu_reference:
        # Special case for translate_to, will be prefixed with 'xx' if not already prefixed with xx because there are only xx.<ISO>.translate_to references
        language = 'xx'
        if nlu_reference[0:3] != 'xx.':
            nlu_reference = 'xx.' + nlu_reference
        logger.info(f'Setting lang as xx for nlu_ref={nlu_reference}')
        component_type = 'translate_to'
    elif len(infos) == 1:
        # if we only have 1 split result, it must a a NLU action reference or an alias
        logger.info('Setting default lang to english')
        language = 'en'
        if infos[0] in nlu.all_components_info.all_components or nlu.all_components_info.all_nlu_actions:
            component_type = infos[0]
    #  check if it is any query of style #<lang>.<class>.<dataset>.<embeddings>
    elif infos[0] in nlu.all_components_info.all_languages:
        language = infos[0]
        component_type = infos[1]
        logger.info(f"Got request for trained model {component_type}")
        if len(infos) == 3:  # dataset specified
            dataset = infos[2]
        if len(infos) == 4:  # embeddings specified
            component_embeddings = infos[3]




    # passing embed_sentence can have format embed_sentence.lang.embedding or embed_sentence.embedding
    # i.e. embed_sentence.bert
    # fr.embed_sentence.bert will automatically select french bert thus no embed_sentence.en.bert or simmilar is required
    # embed_sentence.bert or en.embed_sentence.bert
    # name does not start with a language
    # so query has format <class>.<dataset>
    elif len(infos) == 2:
        logger.info('Setting default lang to english')
        language = 'en'
        component_type = infos[0]
        dataset = infos[1]
    # query has format <class>.<dataset>.<embeddings>
    elif len(infos) == 3:
        logger.info('Setting default lang to english')
        language = 'en'
        component_type = infos[0]
        dataset = infos[1]
        component_embeddings = infos[1]


    logger.info(
        'For input nlu_ref %s detected : \n lang: %s  , component type: %s , component dataset: %s , component embeddings  %s  ',
        nlu_reference, language, component_type, dataset, component_embeddings)
    resolved_component = resolve_component_from_parsed_query_data(language, component_type, dataset,component_embeddings, nlu_reference, trainable,authenticated=authenticated, is_recursive_call=is_recursive_call)
    if resolved_component is None:
        logger.exception("EXCEPTION: Could not create a component for nlu reference=%s", nlu_reference)
        return nlu.NluError()
    return resolved_component


def resolve_component_from_parsed_query_data(language, component_type, dataset, component_embeddings, nlu_ref,trainable=False,path=None,authenticated=False,is_recursive_call=False):
    '''
    Searches the NLU name spaces for a matching NLU reference. From that NLU reference, a SparkNLP reference will be aquired which resolved to a SparkNLP pretrained model or pipeline
    :param nlu_ref: Full request which was passed to nlu.load()
    :param language: parsed language, may never be  '' and should be default 'en'
    :param component_type: parsed component type. may never be ''
    :param dataset: parsed dataset, can be ''
    :param component_embeddings: parsed embeddigns used for the component, can be ''
    :return: returns the nlu.Component class that corrosponds to this component. If it is a pretrained pipeline, it will return a list of components(?)
    '''
    component_kind = ''  # either model or pipe or auto_pipe
    nlp_ref = ''
    logger.info('Searching local Namespaces for SparkNLP reference.. ')
    resolved = False
    is_licensed = False
    # 0. check trainable references
    if trainable == True :
        if nlu_ref in Spellbook.trainable_models.keys():
            component_kind = 'trainable_model'
            nlp_ref = Spellbook.trainable_models[nlu_ref]
            logger.info(f'Found Spark NLP reference in trainable models naqmespace = {nlp_ref}')
            resolved = True
            trainable=True
    # 1. check if pipeline references for resolution
    if resolved == False and language in Spellbook.pretrained_pipe_references.keys():
        if nlu_ref in Spellbook.pretrained_pipe_references[language].keys():
            component_kind = 'pipe'
            nlp_ref = Spellbook.pretrained_pipe_references[language][nlu_ref]
            logger.info(f'Found Spark NLP reference in pretrained pipelines namespace = {nlp_ref}')
            resolved = True

    # 2. check if model references for resolution
    if resolved == False and language in Spellbook.pretrained_models_references.keys():
        if nlu_ref in Spellbook.pretrained_models_references[language].keys():
            component_kind = 'model'
            nlp_ref = Spellbook.pretrained_models_references[language][nlu_ref]
            logger.info(f'Found Spark NLP reference in pretrained models namespace = {nlp_ref}')
            resolved = True

    # 3. check if alias/default references for resolution
    if resolved == False and nlu_ref in Spellbook.component_alias_references.keys():
        sparknlp_data = Spellbook.component_alias_references[nlu_ref]
        component_kind = sparknlp_data[1]
        nlp_ref = sparknlp_data[0]
        logger.info('Found Spark NLP reference in language free aliases namespace')
        resolved = True
        if len(sparknlp_data) > 2 :
            dataset=sparknlp_data[2]
        if len(sparknlp_data) > 3 :
            # special case overwrite for T5
            nlu_ref=sparknlp_data[3]


    # 4. Check Healthcare Pipe Namespace
    if resolved == False and language in Spellbook.pretrained_healthcare_pipe_references.keys():
        if nlu_ref in Spellbook.pretrained_healthcare_pipe_references[language].keys():
            component_kind = 'pipe'
            nlp_ref = Spellbook.pretrained_healthcare_pipe_references[language][nlu_ref]
            logger.info(f'Found Spark NLP reference in pretrained healthcare pipe namespace = {nlp_ref}')
            resolved = True
            is_licensed = True



    # 5. Check Healthcare Model Namespace
    if resolved == False and language in Spellbook.pretrained_healthcare_model_references.keys():
        if nlu_ref in Spellbook.pretrained_healthcare_model_references[language].keys():
            component_kind = 'model'
            nlp_ref = Spellbook.pretrained_healthcare_model_references[language][nlu_ref]
            logger.info(f'Found Spark NLP reference in pretrained healthcare model namespace = {nlp_ref}')
            resolved = True
            is_licensed = True

    # 6. Check Healthcare Aliases Namespace
    if resolved == False and nlu_ref in Spellbook.healthcare_component_alias_references.keys():
        sparknlp_data = Spellbook.healthcare_component_alias_references[nlu_ref]
        component_kind = sparknlp_data[1]
        nlp_ref = sparknlp_data[0]
        logger.info('Found Spark NLP reference in language free healthcare aliases namespace')
        resolved = True
        is_licensed = True
        # if len(sparknlp_data) > 2 :
        #     dataset=sparknlp_data[2]
        # if len(sparknlp_data) > 3 :
        #     # special case overwrite for T5
        #     nlu_ref=sparknlp_data[3]




    # 7. If reference is none of the Namespaces, it must be a component like tokenizer or YAKE or Chunker etc....
    # If it is not, then it does not exist and will be caught later
    if not resolved:
        resolved = True
        component_kind = 'component'
        logger.info('Could not find reference in NLU namespace. Assuming it is a component that is an ragmatic NLP annotator with NO model to download behind it.')


    if not auth_utils.is_authorized_enviroment() and is_licensed :
        print(f"The nlu_ref=[{nlu_ref}] is pointing to a licensed Spark NLP Annotator or Model [{nlp_ref}]. \n"
              f"Your environment does not seem to be Authorized!\n"
              f"Please RESTART your Python environment and run nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)  \n"
              f"with your corrosponding credentials. If you have no credentials you can get a trial version with credentials here https://www.johnsnowlabs.com/spark-nlp-try-free/ \n"
              f"Or contact us at contact@jonsnowlabs.com\n"
              f"NLU will ignore this error and continue running, but you will encounter errors most likely. ")


    if nlp_ref == '' and not is_recursive_call and not trainable and 'en.' not in nlu_ref and language in ['','en']:
        # logger.info('')
        #Search again but with en. prefixed, enables all refs to work withouth en prefix
        # return nlu_ref_to_component(language, component_type, dataset, component_embeddings, 'en.'+nlu_ref,trainable,path,authenticated,True)
        return nlu_ref_to_component('en.'+nlu_ref,is_recursive_call=True)

    # Convert references into NLU Component object which embelishes NLP annotators
    if component_kind == 'pipe':
        constructed_components = construct_component_from_pipe_identifier(language, nlp_ref, nlu_ref,is_licensed=is_licensed)
        logger.info(f'Inferred Spark reference nlp_ref={nlp_ref} and nlu_ref={nlu_ref}  to NLP Annotator Class {constructed_components}')
        if constructed_components is None:
            raise ValueError(f'EXCEPTION : Could not create NLU component for nlp_ref={nlp_ref} and nlu_ref={nlu_ref}  to NLP Annotator Classes {constructed_components}')
        else:
            return constructed_components
    elif component_kind in ['model', 'component']:
        constructed_component = construct_component_from_identifier(language, component_type, dataset,
                                                                    component_embeddings, nlu_ref,
                                                                    nlp_ref,is_licensed=is_licensed)

        logger.info(f'Inferred Spark reference nlp_ref={nlp_ref} and nlu_ref={nlu_ref}  to NLP Annotator Class {constructed_component}')

        if constructed_component is None:
            raise ValueError(f'EXCEPTION : Could not create NLU component for nlp_ref={nlp_ref} and nlu_ref={nlu_ref}')
        else:
            return constructed_component
    elif component_kind == 'trainable_model':
        constructed_component = construct_trainable_component_from_identifier(nlu_ref,nlp_ref)
        if constructed_component is None:
            raise ValueError(f'EXCEPTION : Could not create NLU component for nlp_ref={nlp_ref} and nlu_ref={nlu_ref}')
        else:
            constructed_component.info.is_untrained = True
            return constructed_component
    else:
        raise ValueError(f'EXCEPTION : Could not create NLU component for nlp_ref={nlp_ref} and nlu_ref={nlu_ref}')


def construct_trainable_component_from_identifier(nlu_ref,nlp_ref):
    '''
    This method returns a Spark NLP annotator Approach class embelished by a NLU component
    :param nlu_ref: nlu ref to the trainable model
    :param nlp_ref: nlp ref to the trainable model
    :return: trainable model as a NLU component
    '''

    logger.info(f'Creating trainable NLU component for nlu_ref = {nlu_ref} and nlp_ref = {nlp_ref}')
    try:
        if nlu_ref in ['train.deep_sentence_detector','train.sentence_detector']:
            #no label col but trainable?
            return  nlu.NLUSentenceDetector(annotator_class = 'deep_sentence_detector', trainable='True',nlu_ref=nlu_ref,)
        if nlu_ref in ['train.context_spell','train.spell'] :
            pass
        if nlu_ref in ['train.symmetric_spell'] :
            pass
        if nlu_ref in ['train.norvig_spell'] :
            pass
        if nlu_ref in ['train.unlabeled_dependency_parser'] :
            pass
        if nlu_ref in ['train.labeled_dependency_parser'] :
            pass
        if nlu_ref in ['train.classifier_dl','train.classifier'] :
            return nlu.Classifier(annotator_class = 'classifier_dl', trainable=True, nlu_ref=nlu_ref,)
        if nlu_ref in ['train.ner','train.named_entity_recognizer_dl'] :
            return nlu.Classifier(annotator_class = 'ner', trainable=True,nlu_ref=nlu_ref,)
        if nlu_ref in ['train.sentiment_dl','train.sentiment'] :
            return nlu.Classifier(annotator_class = 'sentiment_dl', trainable=True,nlu_ref=nlu_ref,)
        if nlu_ref in ['train.vivekn_sentiment'] :
            pass
        if nlu_ref in ['train.pos'] :
            return nlu.Classifier(annotator_class = 'pos', trainable=True,nlu_ref=nlu_ref,)
        if nlu_ref in ['train.multi_classifier'] :
            return nlu.Classifier(annotator_class = 'multi_classifier', trainable=True,nlu_ref=nlu_ref,)
        if nlu_ref in ['train.word_seg','train.word_segmenter'] :
            return nlu.Tokenizer(annotator_class = 'word_segmenter', trainable=True,nlu_ref=nlu_ref,)


        if nlu_ref in ['train.generic_classifier'] :
            return nlu.Classifier(annotator_class = 'generic_classifier', trainable=True, nlu_ref=nlu_ref,is_licensed=True)

        if nlu_ref in ['train.resolve_chunks'] :
            return nlu.Resolver(annotator_class = 'chunk_entity_resolver', trainable=True,nlu_ref=nlu_ref,)
        if nlu_ref in ['train.resolve_sentence','train.resolve'] :
            return nlu.Resolver(annotator_class = 'sentence_entity_resolver', trainable=True,nlu_ref=nlu_ref,)

        if nlu_ref in ['train.assertion','train.assertion_dl'] : # TODO
            return nlu.Classifier(annotator_class = 'sentiment_dl', trainable=True,nlu_ref=nlu_ref,)



    except:  # if reference is not in namespace and not a component it will cause a unrecoverable crash
        ValueError(f'EXCEPTION: Could not create trainable NLU component for nlu_ref = {nlu_ref} and nlp_ref = {nlp_ref}')


def construct_component_from_pipe_identifier(language, nlp_ref, nlu_ref,path=None, is_licensed=False, strict=False):
    '''
    # creates a list of components from a Spark NLP Pipeline reference
    # 1. download pipeline
    # 2. unpack pipeline to annotators and create list of nlu components
    # 3. return list of nlu components
    :param nlu_ref:
    :param language: language of the pipeline
    :param nlp_ref: Reference to a spark nlp petrained pipeline
    :param path: Load pipe from HDD
    :return: Each element of the SaprkNLP pipeline wrapped as a NLU componed inside of a list
    '''

    logger.info("Starting Spark NLP to NLU pipeline conversion process")
    from sparknlp.pretrained import PretrainedPipeline, LightPipeline
    if 'language' in nlp_ref: language = 'xx'  # special edge case for lang detectors
    if path == None :
        if is_licensed:
            pipe = PretrainedPipeline(nlp_ref, lang=language,remote_loc='clinical/models')
        else :
            pipe = PretrainedPipeline(nlp_ref, lang=language)
        iterable_stages = pipe.light_model.pipeline_model.stages
    else :
        pipe = LightPipeline(PipelineModel.load(path=path))
        iterable_stages = pipe.pipeline_model.stages
    constructed_components = []

    # for component in pipe.light_model.pipeline_model.stages:
    for component in iterable_stages:

        logger.info(f"Extracting model from Spark NLP pipeline: {component} and creating Component")
        parsed = str(component).split('_')[0].lower()
        logger.info(f"Parsed Component for : {parsed}")
        if isinstance(component, NerConverter):
            constructed_components.append(Util(annotator_class='ner_converter', model=component, lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, MultiClassifierDLModel):
            constructed_components.append(nlu.Classifier(model=component, annotator_class='multi_classifier', lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, PerceptronModel):
            constructed_components.append(nlu.Classifier(annotator_class='pos', model=component, lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, (ClassifierDl,ClassifierDLModel)):
            constructed_components.append(nlu.Classifier(annotator_class='classifier_dl', model=component, lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, UniversalSentenceEncoder):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='use',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, BertEmbeddings):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='bert',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, AlbertEmbeddings):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='albert',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, XlnetEmbeddings):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='xlnet',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, WordEmbeddingsModel):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='glove',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, ElmoEmbeddings):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='elmo',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, BertSentenceEmbeddings):
            constructed_components.append(nlu.Embeddings(model=component, annotator_class='sentence_bert',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True,do_ref_checks=False))
        elif isinstance(component, TokenizerModel):
            constructed_components.append(nlu.Tokenizer(model=component, annotator_class='default_tokenizer',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, DocumentAssembler):
            constructed_components.append(nlu.Util(model=component,lang='xx', nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, SentenceDetectorDLModel):
            constructed_components.append(NLUSentenceDetector(annotator_class='deep_sentence_detector', model=component,lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, SentenceDetector):
            constructed_components.append(NLUSentenceDetector(annotator_class='pragmatic_sentence_detector', model=component,lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif parsed in Spellbook.classifiers:
            constructed_components.append(nlu.Classifier(model=component, lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))


        elif isinstance(component, Chunker):
            constructed_components.append(nlu.chunker.Chunker(annotator_class='default_chunker', model=component,lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, ChunkEmbeddings):
            constructed_components.append(embeddings_chunker.EmbeddingsChunker(annotator_class='chunk_embedder', model=component,lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))



        elif isinstance(component, RegexMatcherModel) or parsed == 'match':
            constructed_components.append(nlu.Matcher(model=component, annotator_class='regex',nlu_ref=nlu_ref))
        elif isinstance(component, TextMatcherModel):
            constructed_components.append(nlu.Matcher(model=component, annotator_class='text',nlu_ref=nlu_ref))
        elif isinstance(component, DateMatcher):
            constructed_components.append(nlu.Matcher(model=component, annotator_class='date',nlu_ref=nlu_ref))
        elif isinstance(component, ContextSpellCheckerModel):
            constructed_components.append(nlu.SpellChecker(model=component, annotator_class='context',nlu_ref=nlu_ref))
        elif isinstance(component, SymmetricDeleteModel):
            constructed_components.append(nlu.SpellChecker(model=component, annotator_class='symmetric',nlu_ref=nlu_ref))
        elif isinstance(component, NorvigSweetingModel):
            constructed_components.append(nlu.SpellChecker(model=component, annotator_class='norvig',nlu_ref=nlu_ref))
        elif isinstance(component, LemmatizerModel):
            constructed_components.append(nlu.lemmatizer.Lemmatizer(model=component,nlu_ref=nlu_ref))
        elif isinstance(component, NormalizerModel):
            constructed_components.append(nlu.normalizer.Normalizer(model=component,nlu_ref=nlu_ref))
        elif isinstance(component, Stemmer):
            constructed_components.append(nlu.stemmer.Stemmer(model=component,nlu_ref=nlu_ref))
        elif isinstance(component, (NerDLModel, NerCrfModel)):
            constructed_components.append(nlu.Classifier(model=component, annotator_class='ner',lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, LanguageDetectorDL):
            constructed_components.append(nlu.Classifier(model=component, annotator_class='language_detector',nlu_ref=nlu_ref))
        elif isinstance(component, DependencyParserModel):
            constructed_components.append(UnlabledDepParser(model=component,nlu_ref=nlu_ref,loaded_from_pretrained_pipe=True ))
        elif isinstance(component, TypedDependencyParserModel):
            constructed_components.append(LabledDepParser(model=component,nlu_ref=nlu_ref,loaded_from_pretrained_pipe=True,))
        elif isinstance(component, MultiClassifierDLModel):
            constructed_components.append(nlu.Classifier(model=component, nlp_ref='multiclassifierdl',nlu_ref=nlu_ref))
        elif isinstance(component, (SentimentDetectorModel,SentimentDLModel)):
            constructed_components.append(nlu.Classifier(model=component, nlp_ref='sentimentdl',nlu_ref=nlu_ref))
        elif isinstance(component, (SentimentDetectorModel,ViveknSentimentModel)):
            constructed_components.append(nlu.Classifier(model=component, nlp_ref='vivekn',nlu_ref=nlu_ref))

        elif isinstance(component, NGram):
            constructed_components.append(nlu.chunker.Chunker(annotator_class='ngram',model=component,nlu_ref=nlu_ref))
        elif isinstance(component, StopWordsCleaner):
            from nlu.components.stopwordscleaner import StopWordsCleaner as Stopw
            constructed_components.append(Stopw(annotator_class='Stopw', model=component,nlu_ref=nlu_ref))
        elif isinstance(component, (TextMatcherModel, RegexMatcherModel, DateMatcher,MultiDateMatcher)) or parsed == 'match':
            constructed_components.append(nlu.Matcher(model=component,nlu_ref=nlu_ref, loaded_from_pretrained_pipe=True))
        elif isinstance(component,(T5Transformer)):
            constructed_components.append(nlu.Seq2Seq(annotator_class='t5', model=component,nlu_ref=nlu_ref))
        elif isinstance(component,(MarianTransformer)):
            constructed_components.append(nlu.Seq2Seq(annotator_class='marian', model=component,nlu_ref=nlu_ref))

        elif isinstance(component, SentenceEmbeddings):
            constructed_components.append(Util(annotator_class='sentence_embeddings', model=component,nlu_ref=nlu_ref))
        elif parsed in Spellbook.word_embeddings + Spellbook.sentence_embeddings:
             constructed_components.append(nlu.Embeddings(model=component, lang=language, nlu_ref=nlu_ref,nlp_ref=nlp_ref,loaded_from_pretrained_pipe=True))
        elif isinstance(component, (Finisher, EmbeddingsFinisher)) : continue # Dont need fnishers since nlu handles finishign
        elif is_licensed :
            from sparknlp_jsl.annotator import AssertionDLModel, AssertionFilterer, AssertionLogRegModel, Chunk2Token, ChunkEntityResolverModel, ChunkFilterer
            from sparknlp_jsl.annotator import ChunkMergeModel, ContextualParserModel, DeIdentificationModel, DocumentLogRegClassifierModel, DrugNormalizer
            from sparknlp_jsl.annotator import GenericClassifierModel, IOBTagger, NerChunker, NerConverterInternal, NerDisambiguatorModel, ReIdentification
            from sparknlp_jsl.annotator import MedicalNerModel,RelationExtractionModel, RelationExtractionDLModel, RENerChunksFilter, SentenceEntityResolverModel
            #todo embelish ChunkFilterer, Chunk2Token, Disambeguiate, DrugNormalizer, RENerChunksfilterer, IOBTager(???)_, ReIdentify,, NERChunker
            if isinstance(component, AssertionLogRegModel):
                constructed_components.append(nlu.Asserter(annotator_class='assertion_dl',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))

            elif isinstance(component, AssertionDLModel):
                constructed_components.append(nlu.Asserter(annotator_class='assertion_log_reg',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))


            elif isinstance(component, SentenceEntityResolverModel):
                constructed_components.append(nlu.Resolver(annotator_class='sentence_entity_resolver',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))

            elif isinstance(component, ChunkEntityResolverModel):
                constructed_components.append(nlu.Resolver(annotator_class='chunk_entity_resolver',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))


            elif isinstance(component, RelationExtractionModel):
                constructed_components.append(nlu.Relation(annotator_class='relation_extractor',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))


            elif isinstance(component, RelationExtractionDLModel):
                constructed_components.append(nlu.Relation(annotator_class='relation_extractor_dl',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))

            elif isinstance(component, MedicalNerModel):
                constructed_components.append(nlu.Classifier(annotator_class='ner_healthcare',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))

            elif isinstance(component, ChunkMergeModel):
                constructed_components.append(nlu.Util(annotator_class='chunk_merger',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))

            elif isinstance(component,ContextualParserModel):
                from nlu.components.chunker import Chunker as Chu
                constructed_components.append(Chu(annotator_class ='contextual_parser',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))


            elif isinstance(component,DeIdentificationModel):
                constructed_components.append(nlu.Deidentification(annotator_class ='deidentifier',model=component,nlu_ref=nlu_ref, nlp_ref=nlp_ref ,loaded_from_pretrained_pipe=True))



            else :
                if strict : raise Exception(f"Could not infer component type for lang={language} and nlp_ref={nlp_ref} of type={component} during pipeline conversion ")
                logger.warning(f"Warning: Could not infer component type for lang={language} and nlp_ref={nlp_ref} and model {component} during pipeline conversion, using default type Normalizer")
                constructed_components.append(nlu.normalizer.Normalizer(model=component))
        else :
            if strict : raise Exception(f"Could not infer component type for lang={language} and nlp_ref={nlp_ref} of type={component} during pipeline conversion ")
            logger.warning(f"Warning: Could not infer component type for lang={language} and nlp_ref={nlp_ref} and model {component} during pipeline conversion, using default type Normalizer")
            constructed_components.append(nlu.normalizer.Normalizer(model=component))

        logger.info(f"Extracted into NLU Component type : {parsed}", )
        if None in constructed_components: raise Exception(f"Could not infer component type for lang={language} and nlp_ref={nlp_ref} during pipeline conversion,")


    return PipeUtils.enforece_AT_embedding_provider_output_col_name_schema_for_list_of_components(ComponentUtils.set_storage_ref_attribute_of_embedding_converters(constructed_components))


def construct_component_from_identifier(language, component_type='', dataset='', component_embeddings='', nlu_ref='',nlp_ref='',is_licensed=False):
    '''
    Creates a NLU component from a pretrained SparkNLP model reference or Class reference.
    Class references will return default pretrained models
    :param language: Language of the sparknlp model reference
    :param component_type: Class which will be used to instantiate the model
    :param dataset: Dataset that the model was trained on
    :param component_embeddings: Embedded that the models was traiend on (if any)
    :param nlu_ref: Full user request
    :param nlp_ref: Full Spark NLP reference
    :return: Returns a NLU component which embelished the Spark NLP pretrained model and class for that model
    '''


    logger.info(f'Creating singular NLU component for type={component_type} sparknlp_ref={nlp_ref} , nlu_ref={nlu_ref} dataset={dataset}, language={language} ' )
    try:
        if 'assert' in component_type:
            return nlu.Asserter(nlp_ref=nlp_ref, nlu_ref=nlu_ref, lang=language, get_default=False, is_licensed=is_licensed)
        elif 'resolve' in component_type or 'resolve' in nlu_ref:
            return nlu.Resolver(nlp_ref=nlp_ref, nlu_ref=nlu_ref, language=language,get_default=False, is_licensed=is_licensed)
        elif 'relation' in nlu_ref:
            return nlu.Relation(nlp_ref=nlp_ref, nlu_ref=nlu_ref, lang=language, get_default=False, is_licensed=is_licensed)
        elif 'de_identify' in nlu_ref and 'ner' not in nlu_ref:
            return nlu.Deidentification(nlp_ref=nlp_ref, nlu_ref=nlu_ref, lang=language, get_default=False, is_licensed=is_licensed)



        elif any(x in Spellbook.seq2seq for x in [nlp_ref, nlu_ref, dataset, component_type, ]):
            return Seq2Seq(annotator_class=component_type, language=language, get_default=False, nlp_ref=nlp_ref,configs=dataset, is_licensed=is_licensed)

        # if any([component_type in NameSpace.word_embeddings,dataset in NameSpace.word_embeddings, nlu_ref in NameSpace.word_embeddings, nlp_ref in NameSpace.word_embeddings]):

        elif any(x in Spellbook.classifiers for x in [nlp_ref, nlu_ref, dataset, component_type, ] + dataset.split('_')):
            return Classifier(get_default=False, nlp_ref=nlp_ref, nlu_ref=nlu_ref, language=language, is_licensed=is_licensed)


        elif any(x in Spellbook.word_embeddings and x not in Spellbook.classifiers for x in
                 [nlp_ref, nlu_ref, dataset, component_type, ] + dataset.split('_')):
            return Embeddings(get_default=False, nlp_ref=nlp_ref, nlu_ref=nlu_ref, lang=language, is_licensed=is_licensed)

        # elif any([component_type in NameSpace.sentence_embeddings,dataset in NameSpace.sentence_embeddings, nlu_ref in NameSpace.sentence_embeddings, nlp_ref in NameSpace.sentence_embeddings]):
        if any(x in Spellbook.sentence_embeddings and not x in Spellbook.classifiers for x in
               [nlp_ref, nlu_ref, dataset, component_type, ] + dataset.split('_')):
            return Embeddings(get_default=False, nlp_ref=nlp_ref, nlu_ref=nlu_ref, lang=language, is_licensed=is_licensed)




        elif any('spell' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return SpellChecker(annotator_class=component_type, language=language, get_default=True, nlp_ref=nlp_ref,nlu_ref=nlu_ref,
                                dataset=dataset, is_licensed=is_licensed)

        elif any('dep' in x and not 'untyped' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return LabledDepParser()

        elif any('dep.untyped' in x or 'untyped' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return UnlabledDepParser()

        elif any('lemma' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.lemmatizer.Lemmatizer(language=language, nlp_ref=nlp_ref, is_licensed=is_licensed)

        elif any('norm' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.normalizer.Normalizer(nlp_ref=nlp_ref, nlu_ref=nlu_ref, is_licensed=is_licensed)
        elif any('clean' in x or 'stopword' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.StopWordsCleaners(lang=language, get_default=False, nlp_ref=nlp_ref)
        elif any('sentence_detector' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return NLUSentenceDetector(nlu_ref=nlu_ref, nlp_ref=nlp_ref, language=language, is_licensed=is_licensed)

        elif any('match' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return Matcher(nlu_ref=nlu_ref, nlp_ref=nlp_ref, is_licensed=is_licensed)


        elif any('tokenize' in x or 'segment_words' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.tokenizer.Tokenizer(nlp_ref=nlp_ref, nlu_ref=nlu_ref, language=language,get_default=False, is_licensed=is_licensed)

        elif any('stem' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.Stemmers()

        # supported in future version with auto embed generation
        # elif any('embed_chunk' in x for x in [nlp_ref, nlu_ref, dataset, component_type] ):
        #     return embeddings_chunker.EmbeddingsChunker()    DrugNormalizer,

        elif any('chunk' in x for x in [nlp_ref, nlu_ref, dataset, component_type]):
            return nlu.chunker.Chunker()
        elif component_type == 'ngram':
            return nlu.chunker.Chunker('ngram')


        logger.exception('EXCEPTION: Could not resolve singular Component for type=%s and nlp_ref=%s and nlu_ref=%s',
                         component_type, nlp_ref, nlu_ref)
        return None
        # raise ValueError
    except:  # if reference is not in namespace and not a component it will cause a unrecoverable crash
        logger.exception('EXCEPTION: Could not resolve singular Component for type=%s and nlp_ref=%s and nlu_ref=%s',
                         component_type, nlp_ref, nlu_ref)
        return None
        # raise ValueError


def extract_classifier_metadata_from_nlu_ref(nlu_ref):
    '''
    Extract classifier and metadataname from nlu reference which is handy for deciding what output column names should be
    Strips lang and action from nlu_ref and returns a list of remaining identifiers, i.e [<classifier_name>,<classifier_dataset>, <additional_classifier_meta>
    :param nlu_ref: nlu reference from which to extra model meta data
    :return: [<modelname>, <dataset>, <more_meta>,]  . For pure actions this will return []
    '''
    model_infos = []
    for e in nlu_ref.split('.'):
        if e in nlu.all_components_info.all_languages or e in nlu.spellbook.Spellbook.actions: continue
        model_infos.append(e)
    return model_infos





import nlu
import requests 

modelhub_json_url = 'https://nlp.johnsnowlabs.com/models.json'
data = requests.get(modelhub_json_url).json()
list_of_missing_refrenes = []

def NLP_ref_to_NLU_ref(nlp_ref,lang) -> str :
    """Resolve a Spark NLP reference to a NLU reference"""
    nlu_namespaces_to_check = [nlu.Spellbook.pretrained_pipe_references, nlu.Spellbook.pretrained_models_references, nlu.Spellbook.pretrained_healthcare_model_references, nlu.Spellbook.licensed_storage_ref_2_nlu_ref , nlu.Spellbook.storage_ref_2_nlu_ref]
    for dict_ in nlu_namespaces_to_check:
        if lang in dict_.keys():
            for reference in dict_[lang]:
                if dict_[lang][reference] == nlp_ref:
                    return reference

def get_missing_NLP_models() -> list:
    for model in data : 
        if NLP_ref_to_NLU_ref(model['name'],model['language']) == None: 
            list_of_missing_refrenes.append([f"https://nlp.johnsnowlabs.com/{model['url']}"])
            return list_of_missing_refrenes

miss = get_missing_NLP_models()

print(miss, sep='\n')



from sparknlp.annotator import *

class SparkNLPUse:
    @staticmethod
    def get_default_model():
        return UniversalSentenceEncoder.pretrained() \
            .setInputCols("sentence") \
            .setOutputCol("use_embeddings")

    @staticmethod
    def get_pretrained_model(name, language):
        return UniversalSentenceEncoder.pretrained(name, language) \
            .setInputCols("sentence") \
            .setOutputCol("use_embeddings")




import nlu

b = nlu.load('bert')
data  = 'I love PYOTRCH + NLU <3'
df = b.predict(data)
# CKL mat means th
np_bert_CKL_mat = df['word_embeddings@small_bert_L2_128_embeddings'].values()
"""
 Tensor Docs https://www.tensorflow.org/guide/tensor 
 Ragged Tensor https://www.tensorflow.org/guide/tensor#ragged_tensors 
 Sparse Tensor https://www.tensorflow.org/guide/tensor#sparse_tensors

Righ now NLU returns NP arrays. One sucy thing is, if you take a column of Word Vectors in 


"""

from keras import backend as K
from keras.layers import Layer
import tensorflow as tf
import tensorflow as tf

class DoubleLinearLayer(tf.keras.layers.Layer):
    def __init__(self, n_units=8):
        super().__init__()
        self.n_units = n_units

    def build(self, input_shape):
        self.weights1 = self.add_weight(
            "weights1",
            shape=(int(input_shape[-1]), self.n_units),
            initializer=tf.keras.initializers.RandomNormal(),
        )
        self.weights2 = self.add_weight(
            "weights2",
            shape=(self.n_units, self.n_units),
            initializer=tf.keras.initializers.RandomNormal(),
        )

    def call(self, inputs):
        x = tf.matmul(inputs, self.weights1)
        return tf.matmul(x, self.weights2)

layer = DoubleLinearLayer()
x = tf.ones((3, 100))
layer(x)


print('yp')
print('yp')
print('yp')
print('yp')
data = 'yo'




import tests.nlu_hc_tests.secrets as sct
import unittest
from nlu import *

class TestViz(unittest.TestCase):

    def test_viz_ner(self):
        pipe = nlu.load('ner.conll',verbose=True)
        data = "Donald Trump and Angela Merkel from Germany don't share many oppinions!"
        """If there are multiple components we chould VIZ from, we need to define what takes prescedence"""
        viz = pipe.viz(data,viz_type='ner')
        print(viz)
        print(viz)
        print(viz)


    def test_viz_dep(self):
        pipe = nlu.load('dep.typed',verbose=True)
        data = "Donald Trump and Angela Merkel from Germany don't share many oppinions!"
        viz = pipe.viz(data,viz_type='dep')
        print(viz)
        print(viz)
        print(viz)


    def test_viz_resolution_chunk(self):
        nlu_ref = 'en.resolve_chunk.icd10cm.neoplasms'
        data = """The patient is a 5-month-old infant who presented initially on Monday with a cold, cough, and runny nose for 2 days. Mom states she had no fever. Her appetite was good but she was spitting up a lot. She had no difficulty breathing and her cough was described as dry and hacky. At that time, physical exam showed a right TM, which was red. Left TM was okay. She was fairly congested but looked happy and playful. She was started on Amoxil and Aldex and we told to recheck in 2 weeks to recheck her ear. Mom returned to clinic again today because she got much worse overnight. She was having difficulty breathing. She was much more congested and her appetite had decreased significantly today. She also spiked a temperature yesterday of 102.6 and always having trouble sleeping secondary to congestion."""

        SPARK_NLP_LICENSE     = sct.SPARK_NLP_LICENSE
        AWS_ACCESS_KEY_ID     = sct.AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY = sct.AWS_SECRET_ACCESS_KEY
        JSL_SECRET            = sct.JSL_SECRET

        nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
        pipe = nlu.load(nlu_ref,verbose=True)
        viz = pipe.viz(data,viz_type='resolution')
        print(viz)


    def test_viz_resolution_sentence(self):
        nlu_ref = 'en.resolve.icd10cm.augmented'
        data = "This is an 82 - year-old male with a history of prior tobacco use , hypertension , chronic renal insufficiency , COPD , gastritis , and TIA who initially presented to Braintree with a non-ST elevation MI and Guaiac positive stools , transferred to St . Margaret\'s Center for Women & Infants for cardiac catheterization with PTCA to mid LAD lesion complicated by hypotension and bradycardia requiring Atropine , IV fluids and transient dopamine possibly secondary to vagal reaction , subsequently transferred to CCU for close monitoring , hemodynamically stable at the time of admission to the CCU ."
        secrets_json_path = '/home/ckl/old_home/Documents/freelance/jsl/tigerGraph/Presentation_2_graph_Summit/secrets.json'
        SPARK_NLP_LICENSE     = sct.SPARK_NLP_LICENSE
        AWS_ACCESS_KEY_ID     = sct.AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY = sct.AWS_SECRET_ACCESS_KEY
        JSL_SECRET            = sct.JSL_SECRET
        nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
        pipe = nlu.load(nlu_ref,verbose=True)
        viz = pipe.viz(data,viz_type='resolution')
        print(viz)
    def test_viz_relation(self):
        nlu_ref = 'med_ner.jsl.wip.clinical relation.temporal_events'
        data = "He was advised chest X-ray or CT scan after checking his SpO2 which was <= 93%"
        secrets_json_path = '/home/ckl/old_home/Documents/freelance/jsl/tigerGraph/Presentation_2_graph_Summit/secrets.json'
        SPARK_NLP_LICENSE     = sct.SPARK_NLP_LICENSE
        AWS_ACCESS_KEY_ID     = sct.AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY = sct.AWS_SECRET_ACCESS_KEY
        JSL_SECRET            = sct.JSL_SECRET
        nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
        pipe = nlu.load(nlu_ref,verbose=True)
        viz = pipe.viz(data,viz_type='relation')
        print(viz)




    def test_viz_assertion(self):
        nlu_ref = 'med_ner.jsl.wip.clinical assert'
        data = "The patient was tested for cancer, but none was detected, he is free of cancer."
        secrets_json_path = '/home/ckl/old_home/Documents/freelance/jsl/tigerGraph/Presentation_2_graph_Summit/secrets.json'
        SPARK_NLP_LICENSE     = sct.SPARK_NLP_LICENSE
        AWS_ACCESS_KEY_ID     = sct.AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY = sct.AWS_SECRET_ACCESS_KEY
        JSL_SECRET            = sct.JSL_SECRET
        nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
        pipe = nlu.load(nlu_ref,verbose=True)
        viz = pipe.viz(data,viz_type='assert')
        print(viz)


    def test_infer_viz_type(self):
        secrets_json_path = '/home/ckl/old_home/Documents/freelance/jsl/tigerGraph/Presentation_2_graph_Summit/secrets.json'
        SPARK_NLP_LICENSE     = sct.SPARK_NLP_LICENSE
        AWS_ACCESS_KEY_ID     = sct.AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY = sct.AWS_SECRET_ACCESS_KEY
        JSL_SECRET            = sct.JSL_SECRET
        nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
        nlu_ref = 'med_ner.jsl.wip.clinical assert'
        data = "The patient was tested for cancer, but none was detected, he is free of cancer."
        pipe = nlu.load(nlu_ref)
        pipe.viz(data)
if __name__ == '__main__':
    unittest.main()




import os,sys, json

import nlu
from nlu.pipe.pipe_components import SparkNLUComponent
# from nlu.pipe.pipe_logic import PipelineQueryVerifier
from nlu.pipe.pipeline import  *
from nlu.pipe.pipe_components import SparkNLUComponent
from pyspark.ml import PipelineModel

from sparknlp.annotator import *

def is_pipe(model_path):
    """Check wether there is a pyspark pipe stored in path"""
    return 'stages' in os.listdir(model_path)
def is_model(model_path):
    """Check wether there is a model stored in path"""
    return'metadata' in os.listdir(model_path)

def get_model_class(model_path):
    """Extract class from a model saved in model_path"""
    with open(model_path+'/stages/part-00000') as json_file:
        java_class = json.load(json_file)['class']
        pyth_class = java_class.split('.')[-1]
    return java_class,pyth_class

def verify_and_create_model(model_path:str, nlu_referenced_requirements = [], nlp_referenced_requirements = [], test_data='Hello world from John Snow Labs '):
    """
     Build model with requirements
     Figures out class name by checking metadata json file
     assumes metadatra is always called part-00000
    """
    with open(model_path+'/metadata/'+'part-00000') as json_f:
        class_name = json.load(json_f)['class'].split('.')[-1]
        # The last element in the Class name can be used to just load the model from disk!
        # Just call eval on it, which will give you the actual Python class reference which should have a .load() method
        try :
            m = eval(class_name).load(model_path)
        except:
            from nlu.utils.environment.offline_load_utils_licensed import verify_model_licensed
            m = verify_model_licensed(class_name, model_path)



    component_type,nlu_anno_class, = resolve_annotator_class_to_nlu_component_info(class_name)
    # Wrap model with NLU Custom Model class so the NLU pipeline Logic knows what to do with it
    c = CustomModel(annotator_class=nlu_anno_class, component_type=component_type, model=m)
    if nlu.hard_offline_checks:
        storage_ref =  m.getStorageRef() if hasattr(m,'getStorageRef')  else -1
        feeds_from_embeddings = False
        for inp in c.info.inputs() :
            if 'embed' in inp : feeds_from_embeddings = True
        if type(m) not in [SentenceEmbeddings,ChunkEmbeddings]  and storage_ref != -1 and feeds_from_embeddings:
            # Check storage ref is matching either some NLU or NLP ref.
            if not check_if_storage_ref_exists(storage_ref) :
                pass
            raise ValueError("When loading a model from disk with embeddings in it's inputs or you must set the Storage Reference attribute on the model."
                             "Additionaly, the configured Storage Reference must either match to a NLU-Model-Reference or a Spark-NLP-Model-Reference."
                             "To view all references visit the modelshub or view https://github.com/JohnSnowLabs/nlu/blob/master/nlu/spellbook.py "
                             "You can set a storage reference via calling model.setStorageRef(storage_reference)"
                             "To disable this check, run nlu.disable_hard_offline_checks()")

    return c
def resolve_annotator_class_to_nlu_component_info(anno_class ='LemmatizerModel'):
    """
    SparkNLUComponent.__init__(self, annotator_class, component_type)
    RECURISIVELY SEARCH through NLU COMPONENTS source code for each class for a given <CLASSS>
    Find the file, which called <CLASS>.pretrained or just <CLASS> !!!
    In the folder containing that found file there will be the component_json info we need!]\
    """
    import nlu
    p = nlu.nlu_package_location+'components/'
    # Check if class has pretrained. If not, dont add .

    import os
    for dirpath, dirs, files in os.walk(p):
        # search for folder that has the component info for that anno_class
        for filename in files:
            if '.py' not in filename or '.pyc' in filename : continue
            fname = os.path.join(dirpath,filename)
            # check annotator name and also getter is in file. This should only be true for the correct annotator creator files
            if (test_check_if_string_in_file(fname, anno_class) and test_check_if_string_in_file(fname, 'def get_.*:', True)) and\
                    (test_check_if_string_in_file(fname, 'return *'+ anno_class, True)):

                parts = dirpath.split('/')
                component_type = parts[-2]
                nlu_anno_class = parts[-1]
                component_type = component_type[:-1]
                return component_type,nlu_anno_class,
    print("COULD NOT FIND COMPONENT INFO FOR ANNO_CLASS", anno_class)
    return False

def test_check_if_string_in_file(file_name, string_to_search, regex=False):
    """ Check if any line in the file contains given string """
    # Open the file in read only mode
    # print('reading ', file_name)
    import re

    with open(file_name, 'r') as read_obj:
        # Read all lines in the file one by one
        for line in read_obj:
            # For each line, check if line contains the string
            if regex :
                    if len(re.findall(string_to_search,line)) > 0 : return True
            else :
                if string_to_search in line: return True
    return False


def NLP_ref_to_NLU_ref(nlp_ref,lang) :
    """Resolve a Spark NLP reference to a NLU reference"""
    nlu_namespaces_to_check = [nlu.Spellbook.pretrained_pipe_references, nlu.Spellbook.pretrained_models_references]
    for dict_ in nlu_namespaces_to_check:
        if lang in dict_.keys():
            for reference in dict_[lang]:
                if dict_[lang][reference] == nlp_ref:
                    return reference


class CustomModel(SparkNLUComponent):
    """ Builds a NLU Components with component info"""
    def __init__(self, annotator_class='sentiment_dl',  component_type='classifier', model = None):
        self.model = model
        SparkNLUComponent.__init__(self, annotator_class, component_type)
        # Make sure input/output cols match up with NLU defaults
        if len(self.info.spark_input_column_names) == 1 :
            model.setInputCols(self.info.spark_input_column_names[0])
        else :
            model.setInputCols(self.info.spark_input_column_names)

        if len(self.info.spark_output_column_names) == 1 :
            model.setOutputCol(self.info.spark_output_column_names[0])
        else :
            model.setOutputCol(self.info.spark_output_column_names)



    # pipe = NLUPipeline()
    # pipe.add(c)
    #
    # # get requirements
    # if len(nlu_referenced_requirements)==0 and len(nlp_referenced_requirements) == 0 :
    #     pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)
    #     res = pipe.predict(test_data)
    #     print(res)
    #     return res
    # elif len(nlu_referenced_requirements) >0:
    #     for r in nlu_referenced_requirements:
    #         pipe.add(nlu.pipe.component_resolution.nlu_ref_to_component(r))
    # elif len(nlp_referenced_requirements) > 0:
    #     for r in nlu_referenced_requirements:
    #         # map back to NLU ref
    #         pipe.add(nlu.pipe.component_resolution.nlu_ref_to_component(NLP_ref_to_NLU_ref(r)))
    # # run pipe with dependencies
    # pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)
    # res = pipe.predict(test_data)
    # print(res)
    # return res
def check_if_storage_ref_exists(storage_ref):
    """Check if storage ref exists as a NLP ref or NLU ref in the spellbook"""
    spaces = [nlu.spellbook.Spellbook.pretrained_models_references, nlu.Spellbook.pretrained_healthcare_model_references]
    for space in spaces :
        if check_namespace_for_storage_ref(space,storage_ref) : return True
    return False
def check_namespace_for_storage_ref(space, storage_ref):
    """Check if storage ref exists as a NLP ref or NLU ref in a given namespace"""
    for lang, mappings in space.items():
        for nlu_ref, nlp_ref in mappings.items():
            if nlu_ref == storage_ref or nlp_ref == storage_ref : return True
    return False





import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class TestE2E(unittest.TestCase):

    def test_e2e_model(self):
        df = nlu.load('en.classify.e2e',verbose=True).predict('You are so stupid', output_level='document')

        for c in df.columns: print(df[c])

        df = nlu.load('e2e',verbose=True).predict('You are so stupid', output_level='sentence')
        for c in df.columns: print(df[c])





if __name__ == '__main__':
    unittest.main()




import nlu
import streamlit as st
from sparknlp.annotator import NerConverter,DependencyParserModel
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from sparknlp.annotator import *
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.discovery import Discoverer
def visualize_classes(
        pipe, # nlu pipe
        text:Union[str,list,pd.DataFrame, pd.Series]='I love NLU and Streamlit and sunny days!',
        output_level:Optional[str]='document',
        title: Optional[str] = "Text Classification",
        metadata : bool = False,
        positions : bool = False,
        set_wide_layout_CSS:bool=True,
        generate_code_sample:bool = False,
        key = "NLU_streamlit"
)->None:

    if title:st.header(title)
    if generate_code_sample: st.code(get_code_for_viz('CLASSES',StreamlitUtilsOS.extract_name(pipe),text))
    classifier_pipes = [pipe]
    classifier_components_usable = [e for e in Discoverer.get_components('classify',True, include_aliases=True)]
    classifier_components = StreamlitUtilsOS.find_all_classifier_components(pipe)
    loaded_classifier_nlu_refs = [c.info.nlu_ref for c in classifier_components]
    # VizUtilsStreamlitOS.classifiers =

    for l in loaded_classifier_nlu_refs:
        if 'converter' in l :
            loaded_classifier_nlu_refs.remove(l)
            continue
        if l not in classifier_components_usable : classifier_components_usable.append(l)

    classifier_components_selection   = st.sidebar.multiselect("Pick additional Classifiers",options=classifier_components_usable,default=loaded_classifier_nlu_refs,key = key)
    classifier_algos_to_load = list(set(classifier_components_selection) - set(loaded_classifier_nlu_refs))
    for embedder in classifier_algos_to_load:classifier_pipes.append(nlu.load(embedder))
    dfs = []
    all_classifier_cols=[]
    for p in classifier_pipes :
        df = p.predict(text, output_level=output_level, metadata=metadata, positions=positions)
        classifier_cols = StreamlitUtilsOS.get_classifier_cols(p)
        for c in classifier_cols :
            if c not in df.columns : classifier_cols.remove(c)

        if 'text' in df.columns: classifier_cols += ['text']
        elif 'document' in df.columns: classifier_cols += ['document']
        all_classifier_cols+= classifier_cols
        dfs.append(df)
    df = pd.concat(dfs, axis=1)
    df = df.loc[:,~df.columns.duplicated()]
    for c in all_classifier_cols :
        if c not in df.columns : all_classifier_cols.remove(c)
    all_classifier_cols = list(set(all_classifier_cols))

    if len(all_classifier_cols) == 0: st.warning('No classes detected')
    else :st.write(df[all_classifier_cols],key=key)


nlu.enable_streamlit_caching()
st.sidebar.write('lol') # elmo albert en.embed.bert.base_cased electra xlnet ar.embed.cbow biobert
visualize_classes(nlu.load('ner '),'hell oworld wahs tis up Peter Licks to FUCK pepeonle')


from nlu.pipe.pipe_components import SparkNLUComponent
class Asserter(SparkNLUComponent):
    def __init__(self, annotator_class='assertion_dl', lang='en', component_type='assertion', get_default=True, model = None, nlp_ref ='', nlu_ref='', trainable=False, is_licensed=False, loaded_from_pretrained_pipe=False):

        if model != None : self.model = model
        else :
            if annotator_class == 'assertion_dl':
                from nlu.components.assertions.assertion_dl.assertion_dl import AssertionDL
                if trainable : self.model = AssertionDL.get_default_trainable_model()
                elif get_default : self.model = AssertionDL.get_default_model()
                else : self.model = AssertionDL.get_pretrained_model(nlp_ref, lang)

            elif annotator_class == 'assertion_log_reg':
                from nlu.components.assertions.assertion_log_reg.assertion_log_reg import AssertionLogReg
                if trainable : self.model = AssertionLogReg.get_default_trainable_model()
                elif get_default : self.model = AssertionLogReg.get_default_model()
                else : self.model = AssertionLogReg\
                    .get_pretrained_model(nlp_ref, lang)


        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, lang,loaded_from_pretrained_pipe , is_licensed)
